# -*- coding: utf-8 -*-
"""Utils Module
"""
import os
import re
from shutil import rmtree
from typing import Tuple
from pathlib import Path
import numpy as np
import pandas as pd
from .fiberstools import getBundleSize, fiber_lens, mesures_to_dataframe
from .deform import deform as _deform
from .sampling import slice_fibers as _sampling
from .intersection import intersection as _intersection
from .postprocessing import postprocessing as _postprocessing


def check_input_path(in_file: str, ext="", dir=False):
    p = Path(in_file)
    if ext and ('.' not in p.name or p.name.rsplit('.', maxsplit=1)[1] != ext):
        raise ValueError(f"File {p.name} must have extension {ext}.")
    if dir and not p.is_dir():
        raise ValueError(f"Path {p} is not a directory.")
    if not p.exists():
        if dir:
            raise FileNotFoundError(f"Directory {p} not found.")
        else:
            raise FileNotFoundError(f"File {p} not found.")

def check_output_path(out_file: str, ext="", dir=False, overwrite=True):
    p = Path(out_file)
    if ext and ('.' not in p.name or p.name.rsplit('.', maxsplit=1)[1] != ext):
        raise ValueError(f"File {p.name} must have extension {ext}.")
    if dir:
        if p.exists():
            if not p.is_dir():
                raise ValueError(f"Path exist {p} but is not a directory.")
        else:
            os.makedirs(str(p))
    else:
        if p.exists() and not overwrite:
            raise ValueError(f"File {p} already exist and will not be deleted.")
        if p.is_dir():
            rmtree(str(p))
        if not p.parent.exists():
            os.makedirs(str(p.parent))


_p1 = re.compile(r'^.*\.bundles$')


def read_bundle(file_in: str, points: int = 0) -> list:
    """
    Read tractography in bundles format.

    Parameters
    ----------
    file_in : str
        Path to read of the tractography dataset file in *bundles* format.

    Returns
    -------
    List where each position contains a NumPy array representing the 3D coordinates of a brain fiber.

    """
    if _p1.match(file_in):
        file_in += 'data'
    with open(file_in, 'rb') as f:
        data = np.fromfile(f, dtype=np.float32)
    if points > 0:
        return data.reshape(-1, points * 3 + 1)[:, 1:].reshape(-1, points, 3)
    else:
        start = 0
        fibers = []
        size = data.size
        u32_data = data.view(np.uint32)
        while start < size:
            f_size = u32_data[start] * 3
            start += 1
            end = start + f_size
            fibers.append(data[start:end].reshape(-1, 3))
            start = end
        return fibers


_MINF_TEMPLATE = """attributes = {{
    'binary' : 1,
    'bundles' : {},
    'byte_order' : 'DCBA',
    'curves_count' : {},
    'data_file_name' : '*.bundlesdata',
    'format' : 'bundles_1.0',
    'space_dimension' : 3
}}
"""


def write_bundle(file_out: str, points: list, buffer_size= 1_048_576)  -> None:
    """
    Write tractography in *bundles* format.

    Parameters
    ----------
    file_out : str
        Path to save of the tractography dataset file in *bundles* format.
    points : list
        List where each position contains a NumPy array representing the 3D coordinates of a brain fibers.

    Returns
    -------
    None

    """
    if not _p1.match(file_out):
        file_out += '.bundles'
    if isinstance(points, np.ndarray) and points.ndim == 3:
        nfibers, npoints, _ = points.shape
        data = np.empty((nfibers, npoints * 3 + 1), np.float32)
        data.view(np.uint32)[:, 0] = npoints
        data[:, 1:] = points.reshape(-1, npoints * 3)
        with open(file_out + 'data', 'wb') as f:
            data.tofile(f)
    else:
        nfibers = len(points)
        buffer = bytearray(buffer_size * 4)
        u32_buffer = memoryview(buffer).cast('I')
        f32_buffer = memoryview(buffer).cast('f')
        start = 0
        with open(file_out + 'data', 'wb') as f:
            for fiber in points:
                assert isinstance(fiber, np.ndarray)
                end = 1 + fiber.size + start
                if end > buffer_size:
                    f.write(buffer[:start * 4])
                    start = 0
                    end = 1 + fiber.size + start
                u32_buffer[start] = fiber.shape[0]
                f32_buffer[start + 1:end] = fiber.flat[:]  # type: ignore
                start = end
            f.write(buffer[:start * 4])
    with open(file_out, 'wt') as f:
        f.write(_MINF_TEMPLATE.format(['points', 0], nfibers))


def deform(deform_file: str, file_in: str, file_out: str) -> None:
    """
    Transforms a tractography file to another space using a non-linear deformation image file.

    Parameters
    ----------
    deform_file : str
        Deformation image (image in NIfTI format containing the deformations).
    file_in : str
        Input tractography dataset in bundles format.
    file_out : str
        Path to the transformed tractography dataset.

    Returns
    -------
    None

    Notes
    -----
    This function generates the following files in the specified directory:

    Tractography : bundles files
        Tractography dataset that has been transformed into the MNI space.

    """
    check_input_path(deform_file, ext='nii')
    check_input_path(file_in)
    check_output_path(file_out, ext='bundles')
    _deform(deform_file, file_in, file_out)


def sampling(file_in: str, file_out: str, npoints: int = 21) -> None:
    """
    Performs a fiber sampling by recalculating their points using a specified number of equidistant points.

    Parameters
    ----------
    file_in : str
        Tractography dataset file in *bundles* format.
    file_out : str
        Path to save the tractography dataset sampled at n equidistant points.
    npoints : str
        Number of sampling points (n). Default: *21*.

    Returns
    -------
    None

    Notes
    -----
    This function generates the following files in the specified directory:

    Tractography : bundles files
        Tractography dataset sampled at n equidistant points in *bundles* format

    """
    check_input_path(file_in, ext="bundles")
    check_output_path(file_out, ext="bundles")
    _sampling(file_in, file_out, npoints)


def intersection (file1_in: str, file2_in: str, dir_out: str,
                  distance_thr: float = 10.0) -> Tuple[float, float]:
    """
    Calculates a similarity measure between two sets of brain fibers (fiber clusters or segmented bundles).
    The similarity measure yields a value between *0* and *100%*.

    Parameters
    ----------
    file1_in : str
        Path of the first fiber bundle.
    file2_in : str
        Path of the second fiber bundle.
    dir_out : str
        Directory name to store all the results generated by the algorithm.
    distance_thr : float
        Distance threshold in *mm* used to consider similar two fibers, default: *10*.

    Returns
    -------
    Tuple : [float, float]
        `[float, float]`, The first value indicates the percentage of intersection of the first set of fibers compared to the second set of fibers,
        and the second value indicates the reverse scenario, intersection of the second set of fibers compared to the first set of fibers.
    
    Notes
    -----
    This function generates three tractography dataset files in the ouput directory, as follows:

    *fiber1-fiber2.bundles/fiber1-fiber2.bundlesdata*: 
      Containing the fibers that are considered similar (or intersecting) for both fascicles.
    *only_fiber1.bundles/only_fiber1.bundlesdata*: 
      Containing the fibers that are only in the first fascicle.
    *only_fiber2.bundles/only_fiber2.bundlesdata*: 
      Containing the fibers that are only in the second fascicle.

    """
    p1 = 0
    p2 = 0

    check_output_path(dir_out, dir=True)
    outfile = os.path.join(dir_out,'distance_matrix.txt')
    check_output_path(outfile)

    _intersection(file1_in, file2_in, outfile)

    ar=open(outfile,'rt')
    t=ar.readlines();ar.close()
    m=len(t)
    n=len(t[0].split('\t')[:-1])

    M = np.zeros((m,n),dtype = 'float32')

    for i in range(m):
        fila=t[i][:-1].split('\t')[:-1]
        for j in range(n):
            M[i,j]=float(fila[j])

    F_under_th=np.where( M < distance_thr) 
        
    xp1=list(set(F_under_th[0]))
    yp2=list(set(F_under_th[1]))

    p1=(len(xp1)/getBundleSize(file1_in))*100 
    p2=(len(yp2)/getBundleSize(file2_in))*100 

    bun1 = np.asarray(read_bundle(file1_in))
    bun2 = np.asarray(read_bundle(file2_in))

    bun1_bun2 = np.concatenate((bun1[xp1], bun2[yp2]), axis=0) 
    write_bundle(os.path.join(dir_out,'fiber1-fiber2.bundles'),list(bun1_bun2))

    ind_bun1 = list(range(len(bun1)))
    ind_bun2 = list(range(len(bun2)))

    xp_sup = list(set(ind_bun1) - set(xp1))
    yp_sup = list(set(ind_bun2) - set(yp2))

    bun1 = np.asarray(read_bundle(file1_in))[xp_sup]
    bun2 = np.asarray(read_bundle(file2_in))[yp_sup]

    write_bundle(os.path.join(dir_out,'only_fiber1.bundles'),list(bun1))
    write_bundle(os.path.join(dir_out,'only_fiber2.bundles'),list(bun2))


    return (p1,p2)


def postprocessing(dir_in: str) -> pd.DataFrame:
    """
    Sets of algorithms that can be applied on the results of clustering and segmentation algorithms.

    Parameters
    ----------
    dir_in : str
        Root directory where all segmentation or clustering algorithm results are stored.

    Returns
    -------
    pd.DataFrame
        `pd.DataFrame`, contains the following list of keys: `size`: number of fibers in the bundle,
        `len`: centroid length per bundle, `intra_mean`: mean intra-bundle Euclidean distance.
    
    """

    if not os.path.exists(dir_in):
        print('Directory does not exist!\nTo execute this function, you must first run an algorithm from the clustering module.')
        
   
    in_centroides=os.path.join(dir_in,'centroids','centroids.bundles')
    in_clusters_directory=os.path.join(dir_in,'final_bundles')

    dir_mensures  =  os.path.join(dir_in, 'outputs', 'mensures')

    if not os.path.exists(dir_mensures):
        os.makedirs(dir_mensures)

    # Fibers Lens and Fibers Size
    txt_lens = ''
    txt_sizes = ''

    centroides=read_bundle(in_centroides)
    for i in range(len(centroides)):
        txt_lens += str(fiber_lens(centroides[i]))+'\n'
        txt_sizes += str(getBundleSize(os.path.join(in_clusters_directory,str(i)+'.bundles')))+'\n'

    with open(os.path.join(dir_mensures, 'lens.txt'), 'w') as file:
        file.write(txt_lens)

    with open(os.path.join(dir_mensures, 'sizes.txt'), 'w') as file:
        file.write(txt_sizes)

    # - Distancia intra-clúster mínima, Distancia intra promedio, Rij y DBindex

    _postprocessing(in_centroides, in_clusters_directory,
                    os.path.join(dir_mensures, 'm'), 21)

    return mesures_to_dataframe(dir_mensures)





'''
to-do:
    readColorFile
'''

import os
import random
from pathlib import Path
from OpenGL import GL
import numpy as np
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files
from .VisualizationBaseObject import VisualizationBaseObject, \
    VisualizationObject, propagateToChildren, drawable, config
from .Shaders import Shader
from .BoundingBox import BoundingBox
from .Tools.utilities import findIntegersMultiplierFor
from .Tools.Quaternion import Quaternion
from .Tools import glm
from .Tools.performance import timeit
from ..FiberVis_core import readBundle, readTrk, applyMatrix, reCalculateNormals, readTck

_b_vs = files('phybers.fibervis') / 'shaders' / 'bundle.vs'
_sfs_fs = files('phybers.fibervis') / 'shaders' / 'standardFragmentShader.fs'
_c_vs = files('phybers.fibervis') / 'shaders' / 'cylinder.vs'
_c_gs = files('phybers.fibervis') / 'shaders' / 'cylinder.gs'
_q_glsl = files('phybers.fibervis') / 'shaders' / 'quaternion.glsl'


class Bundle(VisualizationBaseObject):
    ''' Class for drawing all kind of bundle files (currently implemented .bundle)
    '''

    def __init__(self, sPath, shaderDict, parent):
        ''' Initialize a Bundle object.
        It copies references to the path and the gl program.
        Then it loads the information from the file (points, normals, etc).

        It prepares the color texture, then loads it.

        Sends the information to the GPU buffers.

        Initialize  all parameters.

        The drawable variable is set to True.

        Parameters
        ----------
        sPath : str
            String containing path for the bundle file (.bundles)
        shaderProgram : GLInt
            Reference to the gl program containing all the shaders

        '''

        super().__init__(parent, shaderDict)
        self.nShader = 2
        self.identifier = VisualizationObject.Bundle

        self.path = sPath
        self.fileName = Path(sPath).name

        self.points = None
        self.normals = None
        self.color = None
        self.elements = None
        self.fiberSizes = None

        self.bundlesName = None
        self.bundlesStart = None
        self.curvescount = 0
        self.bundlesInterval = None

        # Read the vertex, fiber index and bundle index (self. points, fibers, bundles)
        self._readFibers()

        # Initialize the values on colorTable
        self.colorTable = None
        self._createColorTable()

        # Initialize of texture id and loading of self.colorTable into the GPU
        self.colorTableTexture = None
        self.validBundleColorTexDims = None
        self._loadColorTexture()

        # Load cylinder shader static data
        self.loadStaticUniforms()

        # Creates VBOs & an EBO, then populates them with the point, normal and color data. The elements data goes into the EBO
        self._loadBuffers()

        # Ready to draw
        self.drawable = True

        dims, center = self.calculateBoundingBoxDimCenter()

        self.boundingbox = BoundingBox(shaderDict, self, dims, center)

        print("Loading ready:\n\t", self.curvescount, " fibers.\n\t", len(self.bundlesName), " bundles.")

    def set_color(self, data):
        self.colorTable[:, :3] = data[0]
        self._loadColorTexture()

    def get_center(self):
        return self.boundingbox.get_center() + self.translateMat[:3, 3].flat

    def get_size(self):
        return self.boundingbox.get_size()

    def cleanOpenGL(self):
        print('Cleaning object: ', self)
        GL.glDeleteVertexArrays(self.nShader, self.vao)

        GL.glDeleteBuffers(3, self.vbo)
        GL.glDeleteBuffers(1, [self.ebo])

        GL.glDeleteTextures([self.colorTableTexture])

        self.clean = True


    def calculateBoundingBoxDimCenter(self):
        # We set the boundingbox parameters
        x = self.points[0::3]
        y = self.points[1::3]
        z = self.points[2::3]

        xmin, xmax = x.min(),x.max()
        ymin, ymax = y.min(),y.max()
        zmin, zmax = z.min(),z.max()

        dims = np.array([xmax-xmin, ymax-ymin, zmax-zmin], dtype=np.float32)
        center = np.array([xmin, ymin, zmin], dtype=np.float32)#+dims/2

        return dims, center

    # @averageTimeit
    @timeit
    def _readFibers(self):
        ''' It reads the information from self.path file.
        Also calculates the normals and color attributes.
        Finally creates the ebo buffer.

        Allowed files: .bundles, .trk

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        extension = Path(self.path).suffix

        # Bundle file
        if extension == '.bundles':
            self._openBundle()

        # Trk file
        elif extension == '.trk':
            self._openTrk()

        # Tck file
        elif extension == '.tck':
            self._openTck()

        # Unsopported file
        else:
            raise TypeError('Unsupported file.', extension)


    def _openBundle(self):
        ''' This function reads the dictionary from the .bundles file.
        It then creates the numpy array for points, normals, color, ebo (elements) and fiberSizes (for bundles with different fiber size).

        Then it calls a C function loaded using Ctypes, passing a pointer to the numpy's data. The arrays are populated in C.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        print('Opening bundle file {}'.format(self.path))


        bunFile = self.path + 'data'
        dataSize = os.path.getsize(bunFile)

        ns = dict()
        with open(self.path) as f:
            exec(f.read(), ns)

        bundlescount = ns[ 'attributes' ][ 'bundles' ]
        self.curvescount = ns[ 'attributes' ][ 'curves_count' ]

        self.bundlesName = bundlescount[::2]
        self.bundlesStart = bundlescount[1::2]

        # self.points = np.empty(dataSize//4-self.curvescount, dtype=np.float32)
        # self.normals = np.empty(dataSize//4-self.curvescount, dtype=np.float32)
        # self.color = np.empty((dataSize//4-self.curvescount)//3, dtype=np.int32)
        # self.elements = np.empty((dataSize//4-self.curvescount)//3+self.curvescount, dtype=np.uint32)
        # self.fiberSizes = np.empty(self.curvescount, dtype=np.int32)

        self.bundlesInterval = np.array(self.bundlesStart+[self.curvescount], dtype=np.int32)

        # cfuncs.readBundleFile(
        #       bunFile.encode('utf-8'),
        #       self.points.ctypes.data,
        #       self.normals.ctypes.data,
        #       self.color.ctypes.data,
        #       self.elements.ctypes.data,
        #       self.fiberSizes.ctypes.data,
        #       self.bundlesInterval.ctypes.data,
        #       self.curvescount,
        #       self.bundlesInterval.size)

        data = readBundle(bunFile,self.curvescount,self.bundlesInterval,self.bundlesInterval.size)
        self.points = data['points']
        self.normals = data['normals']
        self.color = data['color']
        self.elements = data['elements']
        self.fiberSizes = data['fiberSizes']

    def _openTrk(self):
        '''

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        print('Opening trk file {}'.format(self.path))

        data = readTrk(self.path)

        self.points = data['points']
        self.normals = data['normals']
        self.color = data['color']

        self.elements = data['elements']
        self.fiberSizes = data['fiberSizes']

        voxelSize = data['header']['voxel_size']
        vox2RasMat = data['header']['vox_to_ras']

        self.curvescount = data['header']['n_count']
        self.bundlesName = [self.fileName]
        self.bundlesStart = [0]


        # cfuncs.readTrkHeader(self.path.encode('utf-8'),
        #       voxelSize.ctypes.data,
        #       nScalars.ctypes.data,
        #       nProperties.ctypes.data,
        #       vox2RasMat.ctypes.data,
        #       nCount.ctypes.data,
        #       fileSize.ctypes.data,
        #       headerSize.ctypes.data)

        # self.curvescount = nCount[0]
        # self.bundlesName = [self.fileName]
        # self.bundlesStart = [0]

        # size = (fileSize//4-nCount*(nProperties+1))*3//(3+nScalars)


        # self.points = np.empty(size, dtype=np.float32)
        # self.normals = np.empty(size, dtype=np.float32)
        # self.color = np.zeros(size//3, dtype=np.int32)

        # self.elements = np.empty(size//3+self.curvescount, dtype=np.uint32)
        # self.fiberSizes = np.empty(nCount, dtype=np.int32)

        # scalars = np.empty(size//3*nScalars, dtype=np.float32)
        # properties = np.empty(nCount*nProperties, dtype=np.float32)


        # cfuncs.readTrkBody(self.path.encode('utf-8'),
        #       headerSize,
        #       nScalars,
        #       nProperties,
        #       self.points.ctypes.data,
        #       self.normals.ctypes.data,
        #       self.elements.ctypes.data,
        #       self.fiberSizes.ctypes.data,
        #       self.curvescount,
        #       scalars.ctypes.data,
        #       properties.ctypes.data)

        inverseVoxelSize = glm.scaleMatrix([1/voxelSize[0], 1/voxelSize[1], 1/voxelSize[2]])
        trkMat = np.matmul(vox2RasMat, inverseVoxelSize)
        halfVoxelSize = glm.translateMatrix(-voxelSize/2)
        trkMat = np.matmul(trkMat, halfVoxelSize)

        # cfuncs.applyMatrix(self.points.ctypes.data,
        #       self.points.size,
        #       trkMat.ctypes.data)
        applyMatrix(self.points, trkMat)

        normal = np.empty(3, dtype=np.float32)
        normal = self.points[:3] - self.points[3:6]
        normal /= np.linalg.norm(normal)

        if (normal-self.normals[:3]).max() >= 0.01:
            print('reCalculateNormals')

            # cfuncs.reCalculateNormals(
            #   self.points.ctypes.data,
            #   self.normals.ctypes.data,
            #   self.fiberSizes.ctypes.data,
            #   self.curvescount)
            reCalculateNormals(
                            self.points,
                            self.normals,
                            self.fiberSizes,
                            self.curvescount)


    def _openTck(self):

        print('Opening tck file {}'.format(self.path))

        data = readTck(self.path)

        self.curvescount = data['n_count']
        self.bundlesName = [self.fileName]
        self.bundlesStart = [0]

        self.points = data['points']
        self.normals = data['normals']
        self.color = data['color']

        self.elements = data['elements']
        self.fiberSizes = data['fiberSizes']





        # nCount = np.empty(1, dtype=np.int32)
        # fileSize = np.empty(1, dtype=np.uint)
        # headerSize = np.empty(1, dtype=np.int32)

        # cfuncs.readTckHeader(self.path.encode('utf-8'),
        #       nCount.ctypes.data,
        #       fileSize.ctypes.data,
        #       headerSize.ctypes.data)

        # self.curvescount = nCount[0]
        # self.bundlesName = [self.fileName]
        # self.bundlesStart = [0]

        # size = int((fileSize[0]-12-12*self.curvescount)//4)

        # self.points = np.empty(size, dtype=np.float32)
        # self.normals = np.empty(size, dtype=np.float32)
        # self.color = np.zeros(size//3, dtype=np.int32)

        # self.elements = np.empty(size//3+self.curvescount, dtype=np.uint32)
        # self.fiberSizes = np.zeros(nCount, dtype=np.int32)

        # cfuncs.readTckBody(self.path.encode('utf-8'),
        #       headerSize,
        #       self.points.ctypes.data,
        #       self.normals.ctypes.data,
        #       self.elements.ctypes.data,
        #       self.fiberSizes.ctypes.data,
        #       self.curvescount)


        # trkObj = nib.streamlines.load(self.path)

        # self.curvescount = len(trkObj.streamlines)

        # # Just for now, cuz I dont know where to retrive information about bundles in a trk
        # self.bundlesName = [self.fileName]
        # self.bundlesStart = [0]

        # self.fiberSizes = np.array([i.shape[0] for i in trkObj.streamlines], dtype=np.int32)

        # size = self.fiberSizes.sum()
        # print(size*3)
        # self.points = np.concatenate(trkObj.streamlines).ravel()
        # self.normals = np.empty(size*3, dtype=np.float32)
        # self.color = np.empty(size, dtype=np.int32)

        # self.elements = np.empty(size+self.curvescount, dtype=np.uint32)

        # bun = np.array(self.bundlesStart+[self.curvescount], dtype=np.int32)

        # cfuncs.createVBOAndEBOFromPoints(
        #       self.points.ctypes.data,
        #       self.normals.ctypes.data,
        #       self.color.ctypes.data,
        #       self.elements.ctypes.data,
        #       self.fiberSizes.ctypes.data,
        #       bun.ctypes.data,
        #       self.curvescount,
        #       bun.size)




    def readColorFile(self, colorFile):
        pass


    def loadAndApplyMatrix(self, matrixFile):
        extension = Path(matrixFile).suffix

        # numpy matrix file
        if extension == '.npy':
            transform = Bundle.loadNumpyMatrix(matrixFile)

        # brain visa (trm) matrix file
        elif extension == '.trm':
            transform = Bundle.loadTrmMatrix(matrixFile)

        elif extension == '.txt':
            transform = Bundle.loadTxtMatrix(matrixFile)

        # Unsopported file
        else:
            raise TypeError('Unsupported matrix file.', extension)

        self.applyMatrix(transform)


    def applyMatrix(self, transform):
        applyMatrix(self.points,transform)
        # cfuncs.applyMatrix(self.points.ctypes.data,
        #       self.points.size,
        #       transform.ctypes.data)

        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[0])
        GL.glBufferSubData(GL.GL_ARRAY_BUFFER, 0, self.points.nbytes, self.points)

        normal = np.empty(3, dtype=np.float32)
        normal = self.points[:3] - self.points[3:6]
        normal /= np.linalg.norm(normal)

        if (normal-self.normals[:3]).max() >= 0.01:
            reCalculateNormals(
                            self.points,
                            self.normals,
                            self.fiberSizes,
                            self.curvescount)
            # cfuncs.reCalculateNormals(
            #   self.points.ctypes.data,
            #   self.normals.ctypes.data,
            #   self.fiberSizes.ctypes.data,
            #   self.curvescount)

            GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[1])
            GL.glBufferSubData(GL.GL_ARRAY_BUFFER,
                0,
                self.normals.nbytes,
                self.normals)

        dims, center = self.calculateBoundingBoxDimCenter()
        self.boundingbox.updateBBModel(dims, center)

    @staticmethod
    def loadNumpyMatrix(numpyMatrixFile):
        return np.load(numpyMatrixFile).astype(np.float32)

    @staticmethod
    def loadTrmMatrix(trmMatrixFile):
        transform = np.zeros(16, dtype=np.float32)

        with open(trmMatrixFile) as f:
            lines = f.readlines()
            translate = lines[0].split(' ')

            transform[0:3] = [float(x) for x in lines[1].split(' ')]
            transform[3] = float(translate[0])

            transform[4:7] = [float(x) for x in lines[2].split(' ')]
            transform[7] = float(translate[1])

            transform[8:11] = [float(x) for x in lines[3].split(' ')]
            transform[11] = float(translate[2])

            transform[15] = 1.0

        return transform

    @staticmethod
    def loadTxtMatrix(txtMatrixFile):
        return np.loadtxt(txtMatrixFile, dtype=np.float32, usecols=range(4))


    def _createColorTable(self):
        ''' It creates a color (rgba) randomly for each bundle in the file.
        The table is then loaded to the GPU as 1D texture.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        self.colorTable = np.empty((len(self.bundlesStart), 4), dtype=np.float32)

        for i in range(len(self.bundlesStart)):
            self.colorTable[i,:] = [random.random(), random.random(), random.random(), 1.0]

        hieTable = [
            # ('AR_LEFT', [233, 18, 0]),
            # ('AR_ANT_LEFT', [65, 149, 45]),
            # ('AR_POST_LEFT', [226, 226, 0]),
            # ('AR_RIGHT', [233, 18, 0]),
            # ('AR_ANT_RIGHT', [65, 149, 45]),
            # ('AR_POST_RIGHT', [226, 226, 0]),
            # ('CST_LEFT', [228, 122, 0]),
            # ('CST_RIGHT', [228, 122, 0]),
            # ('CST_long_LEFT', [228, 122, 0]),
            # ('CST_long_RIGHT', [228, 122, 0]),
            # ('THAL_FRONT_LEFT', [186, 185, 167]),
            # ('THAL_MOT_LEFT', [45, 151, 131]),
            # ('THAL_PAR_LEFT', [237, 172, 167]),
            # ('THAL_TEMP_LEFT', [217, 156, 0]),
            # ('THAL_OCC_LEFT', [143, 201, 199]),
            # ('THAL_FRONT_RIGHT', [186, 185, 167]),
            # ('THAL_MOT_RIGHT', [45, 151, 131]),
            # ('THAL_PAR_RIGHT', [237, 172, 167]),
            # ('THAL_TEMP_RIGHT', [217, 156, 0]),
            # ('THAL_OCC_RIGHT', [143, 201, 199]),
            # ('IFO_LEFT', [226, 108, 232]),
            # ('IFO_RIGHT', [226, 108, 232]),
            # ('IL_LEFT', [103, 51, 90]),
            # ('IL_RIGHT', [103, 51, 90]),
            # ('UN_LEFT', [76, 255, 255]),
            # ('UN_RIGHT', [76, 255, 255]),
            # ('CG2_LEFT', [145, 235, 9]),
            # ('CG_LEFT', [172, 70, 32]),
            # ('CG3_LEFT', [27, 108, 183]),
            # ('CG2_RIGHT', [145, 235, 9]),
            # ('CG_RIGHT', [172, 70, 32]),
            # ('CG3_RIGHT', [27, 108, 183]),
            # ('FORNIX_LEFT', [0, 0, 0]),
            # ('FORNIX_RIGHT', [0, 0, 0]),
            # ('CC_ROSTRUM', [204, 0, 97]),
            # ('CC_SPLENIUM', [75, 39, 32]),
            # ('CC_BODY', [0, 67, 0]),
            # ('CC_GENU', [36, 0, 130]),
            # ('THAL_PAR_LEFT', [237, 162, 147])

            # fibras largas
            ####################################
            # fibras cortas

            ('lh_PoC-PrC_0', [255, 0, 0]),
            ('rh_PoC-PrC_0', [255, 0, 0]),
            ('lh_PoC-PrC_1', [0, 255, 0]),
            ('rh_PoC-PrC_1', [0, 255, 0]),
            ('lh_PoC-PrC_2', [0, 0, 200]),
            ('rh_PoC-PrC_2', [0, 0, 200]),
            ('lh_PoC-PrC_3', [255, 255, 20]),
            ('rh_CAC-PoCi_0', [188, 243, 243]),
            ('rh_PrC-SP_0', [190, 21, 216]),
            ('lh_PoCi-RAC_0', [179, 147, 47]),
            ('rh_PoCi-RAC_0', [179, 147, 47]),
            ('lh_MOF-ST_0', [249, 182, 249]),
            ('rh_MOF-ST_0', [249, 182, 249]),
            ('rh_PrC-SM_0', [125, 100, 50]),
            ('lh_PrC-SM_0', [125, 100, 50]),
            ('lh_Or-Ins_0', [147, 147, 179]),
            ('rh_Or-Ins_0', [147, 147, 179]),
            ('lh_Op-PrC_0', [255, 0, 255]),
            ('rh_Op-PrC_0', [255, 0, 255]),
            ('lh_PrC-SF_0', [250, 100, 0]),
            ('lh_RMF-SF_0', [100, 25, 150]),
            ('rh_RMF-SF_0', [100, 25, 150]),
            ('lh_RMF-SF_1', [80, 230, 180]),
            ('rh_RMF-SF_1', [80, 230, 180]),
            ('lh_LOF-ST_0', [240, 100, 100]),
            ('rh_LOF-ST_0', [240, 100, 100]),
            ('lh_Op-Ins_0', [50, 125, 100]),
            ('rh_Op-Ins_0', [50, 125, 100]),
            ('lh_LOF-RMF_0', [40, 40, 255]),
            ('rh_LOF-RMF_0', [40, 40, 255]),
            ('lh_LOF-RMF_1', [250, 160, 0]),
            ('rh_LOF-RMF_1', [250, 160, 0]),
            ('lh_IP-MT_0', [255, 20, 147]),
            ('rh_IP-MT_0', [255, 20, 147]),
            ('lh_ST-TT_0', [138, 43, 226]),
            ('rh_ST-TT_0', [138, 43, 226]),
            ('lh_CMF-Op_0', [210, 105, 30]),
            ('lh_CAC-PrCu_0', [0, 0, 139]),
            ('rh_CAC-PrCu_0', [0, 0, 139]),
            ('lh_IC-PrCu_0', [0, 150, 0]),
            ('rh_IC-PrCu_0', [0, 150, 0]),
            ('lh_Op-SF_0', [150, 25, 150]),
            ('rh_Op-SF_0', [150, 25, 150]),
            ('lh_PoCi-PrCu_0', [150, 25, 150]),
            ('lh_PoCi-PrCu_1', [180, 147, 180]),
            ('rh_PoCi-PrCu_1', [180, 147, 180]),
            ('rh_PoCi-PrCu_2', [0, 250, 0]),
            ('lh_LOF-Or_0', [255, 255, 20]),
            ('lh_CMF-RMF_0', [50, 100, 125]),
            ('rh_CMF-RMF_0', [50, 100, 125]),
            ('lh_Fu-LO_0', [188, 143, 143]),
            ('rh_Fu-LO_1', [229, 229, 112]),
            ('lh_PoC-Ins_0', [119, 200, 0]),
            ('rh_LO-SP_0', [30, 144, 255]),
            ('lh_Tr-SF_0', [255, 117, 20]),
            ('rh_Tr-SF_0', [255, 117, 20]),
            ('lh_SM-Ins_0', [100, 100, 200]),
            ('rh_SM-Ins_0', [100, 100, 200]),
            ('lh_IP-IT_0', [100, 0, 100]),
            ('rh_IP-IT_0', [100, 0, 100]),
            ('rh_Cu-Li_0', [184, 134, 11]),
            ('lh_CMF-PoC_0', [141, 188, 143]),
            ('rh_Op-Tr_0', [173, 255, 47]),
            ('lh_Tr-Ins_0', [255, 160, 122]),
            ('rh_Tr-Ins_0', [255, 160, 122]),
            ('lh_PrC-Ins_0', [32, 178, 170]),
            ('rh_PrC-Ins_0', [32, 178, 170]),
            ('lh_CMF-SF_0', [224, 176, 255]),
            ('rh_CMF-SF_0', [224, 176, 255]),
            ('rh_CMF-SF_1', [107, 142, 35]),
            ('rh_LOF-MOF_0', [15, 15, 15]),
            ('lh_MT-SM_0', [139, 0, 0]),
            ('rh_MT-SM_0', [139, 0, 0]),
            ('lh_PoC-SM_0', [219, 112, 147]),
            ('rh_PoC-SM_0', [219, 112, 147]),
            ('lh_PoC-SM_1', [152, 251, 152]),
            ('rh_RAC-SF_0', [238, 232, 170]),
            ('lh_RAC-SF_1', [218, 112, 214]),
            ('lh_PoCi-SF_0', [46, 139, 87]),
            ('lh_IP-SM_0', [160, 82, 45]),
            ('rh_IP-SM_0', [160, 82, 45]),
            ('lh_IP-SP_0', [0, 255, 127]),
            ('rh_IP-SP_0', [0, 255, 127]),
            ('lh_IP-SP_1', [244, 164, 96]),
            ('rh_IP-LO_0', [106, 90, 205]),
            ('lh_IP-LO_1', [192, 192, 192]),
            ('rh_PoC-SP_0', [255, 69, 0]),
            ('rh_PoC-SP_1', [72, 209, 204]),
            ('lh_IT-MT_0', [210, 180, 140]),
            ('rh_IT-MT_1', [135, 206, 235]),
            ('rh_IT-MT_2', [199, 21, 133]),
            ('lh_ST-Ins_0', [255, 218, 185]),
            ('lh_SP-SM_0', [128, 0, 0]),
            ('rh_SP-SM_0', [128, 0, 0]),
            ('lh_MT-ST_0', [218, 112, 214]),
            ('rh_MT-ST_0', [218, 112, 214]),
            ('lh_CMF-PrC_0', [205, 92, 92]),
            ('rh_CMF-PrC_0', [205, 92, 92]),
            ('lh_CMF-PrC_1', [205, 133, 63]),
            ('rh_CMF-PrC_1', [205, 133, 63])
        ]

        for i in range(len(self.bundlesName)):
            try:
                self.colorTable[i,:3] = hieTable[[i[0] for i in hieTable].index(self.bundlesName[i])][1]
                self.colorTable[i,:3] /= 255
            except:
                continue



    def _loadBuffers(self):
        ''' It generates a VAO, VBO and EBO.
        It populates them with the vertex, normal and colors (the VBO), also the elements (the EBO).

        It finishes when the buffers are linked with attributes on the vertex shader.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        # Reference for vbos, ebos and attribute ptrs
        self.vao = GL.glGenVertexArrays(self.nShader)
        self.vbo = GL.glGenBuffers(3)
        self.ebo = GL.glGenBuffers(1)



        #################
        # Lines shader
        GL.glBindVertexArray(self.vao[0])

        # Enable attributes
        positionAttribute =     self.shader[0].attributeLocation('vertexPos')
        normalAttribute =       self.shader[0].attributeLocation('vertexNor')
        colorAttribute =        self.shader[0].attributeLocation('vertexCol')

        # VBO
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[0])
        GL.glBufferData(GL.GL_ARRAY_BUFFER, self.points.nbytes, self.points, GL.GL_STATIC_DRAW) # Create empty buffer
        GL.glEnableVertexAttribArray(positionAttribute)
        GL.glVertexAttribPointer(positionAttribute,3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)

        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[1])
        GL.glBufferData(GL.GL_ARRAY_BUFFER, self.normals.nbytes, self.normals, GL.GL_STATIC_DRAW)       # Create empty buffer
        GL.glEnableVertexAttribArray(normalAttribute)
        GL.glVertexAttribPointer(normalAttribute,       3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)

        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[2])
        GL.glBufferData(GL.GL_ARRAY_BUFFER, self.color.nbytes, self.color, GL.GL_STATIC_DRAW)   # Create empty buffer
        GL.glEnableVertexAttribArray(colorAttribute)
        GL.glVertexAttribPointer(colorAttribute,        1, GL.GL_INT, GL.GL_FALSE, 0, None)

        # EBO
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.ebo)
        GL.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, self.elements.nbytes, self.elements, GL.GL_STATIC_DRAW)


        #################
        # Cylinder shader
        GL.glBindVertexArray(self.vao[1])

        # Enable attributes
        positionAttribute =     self.shader[1].attributeLocation('vertexPos')
        colorAttribute =        self.shader[1].attributeLocation('vertexCol')

        # VBO
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[0])
        GL.glEnableVertexAttribArray(positionAttribute)
        GL.glVertexAttribPointer(positionAttribute, 3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)

        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[2])
        GL.glEnableVertexAttribArray(colorAttribute)
        GL.glVertexAttribPointer(colorAttribute, 1, GL.GL_INT, GL.GL_FALSE, 0, None)

        # EBO
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.ebo)

        # Dereference vao
        GL.glBindVertexArray(0)


    def _loadColorTexture(self):
        ''' It generates a texture (if not already). Then makes the texture0 the active one and loads the color table as a 1D texture.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''
        maxTexDim = GL.glGetIntegerv(GL.GL_MAX_TEXTURE_SIZE)

        if maxTexDim*maxTexDim < len(self.bundlesStart):
            raise ValueError('To many bundles inside the file... Color texture not big enought.')
        self.validBundleColorTexDims = findIntegersMultiplierFor(len(self.bundlesStart), maxTexDim)
        # self.validBundleColorTexDims = (maxTexDim, int(math.ceil(len(self.bundlesStart)/maxTexDim)))

        if self.colorTableTexture == None:
            self.colorTableTexture = GL.glGenTextures(1)

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.colorTableTexture)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_BORDER)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_BORDER)

        bgColor = np.array([1.0, 1.0, 1.0, 1.0], dtype=np.float32)
        GL.glTexParameterfv(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_BORDER_COLOR, bgColor)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)

        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, self.validBundleColorTexDims[0], self.validBundleColorTexDims[1], 0, GL.GL_RGBA, GL.GL_FLOAT, self.colorTable.flatten())


    def loadStaticUniforms(self):
        def createCylinder(nFaces, r):
            rotationX = Quaternion.fromAngleAxis(2*np.pi/nFaces, [1, 0, 0])

            data = np.empty((nFaces+1)*2, dtype=np.float32)
            cylinderOffset = np.array([0, cylinderRadius, 0], dtype=np.float32)

            # Rotate vertex
            for i in range(nFaces+1):
                cylinderOffset = rotationX.rotateVector(cylinderOffset)
                data[ 2*i ] = cylinderOffset[1]
                data[2*i+1] = cylinderOffset[2]

            return data

        # create static data
        cylinderNFaces = 7
        cylinderRadius = 0.15
        cylinderData = createCylinder(cylinderNFaces, cylinderRadius);

        # load data
        self.shader[1].glUseProgram()
        GL.glUniform2fv(self.shader[1].glGetUniformLocation("cylinderVertex"), cylinderData.size, cylinderData);


    def loadUniforms(self):
        '''  This function is called always before drawing. It makes sure that the current bundle will be visualized with the uniforms specified for it.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation("M"), 1, GL.GL_TRUE, self.model.getA())
        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation("colorTable"), 0)

        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('texture1DMax'), self.validBundleColorTexDims[0])


    @propagateToChildren
    @drawable
    @config
    def draw(self):
        '''  It has 2 decorators configuring the vao and seeing if the visualize flag is up.

        First it makes active the texture, and binds the texture from this file.
        Finally it calls the draw element with primitive restart.
        '''

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.colorTableTexture)

        if self.selectedShader == 0:
            GL.glDrawElements(GL.GL_LINE_STRIP, self.elements.size, GL.GL_UNSIGNED_INT, None)
        else:
            GL.glEnable(GL.GL_CULL_FACE)
            GL.glCullFace(GL.GL_FRONT);
            GL.glFrontFace(GL.GL_CCW);
            GL.glDrawElements(GL.GL_LINE_STRIP, self.elements.size, GL.GL_UNSIGNED_INT, None);
            GL.glDisable(GL.GL_CULL_FACE);

        self.boundingbox.draw()

    @staticmethod
    def createProgram():
        ''' Anonymous function.
        It creates the shader programs for this specific class and returns the handler.
        '''
        lineVertexGLSL = [str(_b_vs)]
        lineFragmentGLSL = [str(_sfs_fs)]
        cylinderVertexGLSL = [str(_c_vs)]
        cylinderGeometryGLSL = [str(_c_gs), str(_q_glsl)]
        cylinderFragmentGLSL = [str(_sfs_fs)]

        return [Shader(lineVertexGLSL, lineFragmentGLSL),
                Shader(cylinderVertexGLSL, cylinderFragmentGLSL, cylinderGeometryGLSL)]

    @staticmethod
    def validExtension():
        return ['bundles', 'trk', 'tck']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import numpy as N
import pandas as pd
import regex

#%%
_curves_count_p = regex.compile(r"\s*'curves_count'\s*:\s*(\d+),")
def getBundleSize( bundlefile ):
    """ 
    Get Bundles Size from file.
    """
    #get center names from bundle file
    with open(bundlefile, "rt") as f:
        m = _curves_count_p.search(f.read())
        if m:
            return int(m.group(1))
        else:
            raise Exception

def fiber_lens(fiber):
    """
    Find the length of a fiber.
    """
    final_len = 0
    for i in range(fiber.shape[0]-1):
        current_lens = N.linalg.norm(fiber[i,:] - fiber[i+1,:])
        final_len+=current_lens
    return final_len


def mesures_to_dataframe (dirstad):
    """
    Create Dataframe from measures: sizes, lens, intra_means,intra_min and Ri_dbindex.
    """
    tam  = []
    with open(os.path.join(dirstad, "sizes.txt"), "r") as file:
        for i in file:
            tam.append(int(i[:-1]))

    larg  = []
    with open(os.path.join(dirstad,"lens.txt"), "r") as file:
        for i in file:
            larg.append(round(float(i[:-1]),2))


    intra_mean  = []
    with open(os.path.join(dirstad,"m_intra_means_dists.txt"), "r") as file:
        for i in file:
            intra_mean.append(round(float(i.split(" ")[1][:-1]),2))

    # Ri_db  = []
    # with open(os.path.join(dirstad, "m_Ri_db.txt"), "r") as file:
    #     for i in file:
    #         Ri_db.append(round(float(i.split(" ")[1][:-1]),2))


    # data= {'sizes':tam,'lens':larg,'intra_mean':intra_mean,'Ri_db':Ri_db }

    # df_data = pd.DataFrame(data, columns = ['size','len','intra_mean','Ri_db'])

    data= {'size':tam,'len':larg,'intra_mean':intra_mean}

    df_data = pd.DataFrame(data, columns = ['size','len','intra_mean'])

    return df_data



# -*- coding: utf-8 -*-
import sys, os
import numpy as N
from time import time
from scipy import sparse



#class with Average Link Analisys vars
class WeightedGraph():


    def __init__( self, V, edges, weights ):

        self.V = V
        self.E = len( edges )
        self.edges = edges
        self.weights = weights


def createfffGraphFromFile( graphfile ):

    #read graph file and create fff graph
    inFile = open( graphfile, 'r' )
    line = inFile.readline()
    l = line.split()
    V = int( l[ 0 ] )
    E = int( l[ 1 ] )
    edges = N.zeros( [ E, 2 ], 'i' )
    weights = N.zeros( [ E, 1 ], 'f' )
    #print ('edges: ', E, ' vertices: ', V)

    count = 0
    while True:

        line = inFile.readline()
        if line == '': break
        l = line.split()
        edges[ count, 0 ] = int( l[ 0 ] )
        edges[ count, 1 ] = int( l[ 1 ] )
        weights[ count ] = float( l[ 2 ] )
        count += 1
    inFile.close()
    G = WeightedGraph( V, edges, weights )
    return G


#class with Average Link Analisys vars
class CenterWForestVars():


    def __init__( self, wforestFile ):

        f = open( wforestFile, 'r' )

        #number of Wforest nodes
        self.V = int( f.readline() )
        #list of node father
        self.fa = []
        #list of node child indexes (two)
        self.ch = []

        for v in range( self.V ):

            line = f.readline()
            line = line.split()
            self.fa.append( int( line[ 0 ] ) )
            child = [ int( line[ 1 ] ), int( line[ 2 ] ) ]
            self.ch.append( child )

        f.close()
        #number of white matter parcels
        self.Vg = 0
        #points not included in the Weighted Forest (without connectivity)
        self.singles = []
        #number of parcels included in the Weighted Forest (clusterized)
        self.VV = 0
        #number of trees
        self.top = []
        #final top nodes of partition
        self.finalcl = []
        #singles additional nodes after partition
        self.singles2 = []
        #final clusters
        self.clusters = []

    def get_descendents( self, v ):

        v = int( v )
        if ( v >= 0 and v <= self.V - 1 ):

            if ( self.ch[ v ][ 0 ] == -1 and self.ch[ v ][ 1 ] == -1 ):

                return [ v ]

            else:

                descendent = [ v ]

                for c in self.ch[ v ]:

                    temp = self.get_descendents( c )

                    for q in temp:

                        descendent.append( q )

            return descendent


##for clusterizeCentersMaxdist.py
def createSparceScipyMatrixFromfffGraph( graphfile ):

    G = createfffGraphFromFile( graphfile )
    V = G.V
    edges = G.edges
    weights = G.weights
    mat = sparse.lil_matrix( ( V, V ) )
    for ind in range( len( edges ) ):

        ed = edges[ ind ]
        i = ed[ 0 ]
        j = ed[ 1 ]
        mat[ i,j ] = weights[ ind ].item()
    return mat


#calculate max and mean distances within cluster nodes
#(clusters are connected components of the tree and not singles)
def calculate_top_dist( WFV, data, top, maxdist ):

    #number of default clusters
    #(connected components of the tree and not singles)
    defcl = len( top )
    dist_data  = N.zeros( defcl ) # max
    for i in range( defcl ):

        t = top[ i ]
        maxd = calculate_node_dist( WFV, data, t, maxdist )
        dist_data[ i ] = maxd
    return dist_data

#calculate max and mean distances within cluster nodes
#(clusters are connected components of the tree and not singles)
def calculate_top_dist_nomatrix( WFV, sparse_mat, top, minaff ):

    #number of default clusters
    #(connected components of the tree and not singles)
    defcl = len( top )
    dist_data  = N.zeros( defcl ) # max
    for i in range( defcl ):

        t = top[ i ]
        mind = calculate_node_dist_nomatrix( WFV, sparse_mat, t, minaff )
        dist_data[ i ] = mind
    return dist_data

#calculate max and mean distances within cluster nodes
#(clusters are connected components of the tree and not singles)
def calculate_node_dist( WFV, data, node, maxdist ):

    maxd = 0.0
    if ( WFV.ch[ node ][ 0 ] != -1 and WFV.ch[ node ][ 1 ] != -1 ):

        #get all cluster nodes
        s = WFV.get_descendents( node )
        s.remove( node )
        ss = N.array( s )
        #identify leaves
        ind_s = N.where( ss < WFV.Vg )[ 0 ]
        s = ss[ ind_s ]
        for s1 in range( len( s ) ):

            a = s[ s1 ]
            for s2 in range( s1 + 1, len( s ) ):

                b = s[ s2 ]
                val = data[ a ][ b ]
                if val > maxd:

                    maxd = val
                if maxd > maxdist:

                    return val
    return maxd

#(clusters are connected components of the tree and not singles)
def calculate_node_dist_nomatrix( WFV, sparse_mat, node, minaff ):

    #node = WFV.top(1)
    mind = 100.0
    if ( WFV.ch[ node ][ 0 ] != -1 and WFV.ch[ node ][ 1 ] != -1 ):

        #get all cluster nodes
        s = WFV.get_descendents( node )
        s.remove( node )
        ss = N.array( s )
        #identify leaves
        ind_s = N.where( ss < WFV.Vg )[ 0 ]
        s = ss[ ind_s ]
        for s1 in range( len( s ) ):

            a = s[ s1 ]
            #get all edges indexes who have the node
            for s2 in range( s1 + 1, len( s ) ):

                b = s[ s2 ]
                val = sparse_mat[ a, b ]
                if val < mind:

                    mind = val
                if mind <= minaff:

                    return mind
    return mind


def create_partition_from_finalcl( WFV, finalcl ):

    clusters = []
    for f in finalcl:

        if ( WFV.ch[ f ][ 0 ] != -1 and WFV.ch[ f ][ 1 ] != -1 ): #not single

            s = WFV.get_descendents( f )
            s.remove( f )
            ss = N.array( s )
            #identify leaves
            ind_s = N.where( ss < WFV.Vg )[ 0 ]
            s = ss[ ind_s ]
            nodelist = list( s )
        else:

            nodelist = [ f ]
        clusters.append( nodelist )

    WFV.clusters = clusters

def wforest_partition_maxdist_from_graph( wforestfile,
                                       maxdist,
                                       add_singles = False,
                                       graphfile = '',
                                       var: float = 3600 ):

    WFV = CenterWForestVars( wforestfile )
    nomatrix=True
    #number of vertices of the graph (fiber centers)
    Vg = 0
    while ( Vg < len( WFV.ch ) and \
                           WFV.ch[ Vg ][ 0 ] == -1 and WFV.ch[ Vg ][ 1 ] == -1  ):

        Vg = Vg + 1

    singles = []
    for v in range( Vg ):

        if WFV.fa[ v ] == v:
            singles.append( v )
    #
    #number of centers included in the Weighted Forest (clusterized)
    VV = Vg - len( singles )
    #
    #number of trees
    top = []
    for v in range( Vg, WFV.V ):

        if WFV.fa[ v ] == v:

            top.append( v )
    #store WFV data
    WFV.Vg = Vg
    WFV.singles = singles
    WFV.VV = VV
    WFV.top = top

    #calculate maximum values of distances to evaluate clusters
    max_cldist = float(maxdist) * 1.0         #maximum distance within a cluster
    max_meancldist = float(maxdist) * 1.0      #mean distance within a cluster
    #minimum affinity within a cluster
    min_claff = N.exp( -max_cldist * max_cldist / float(var) );
    #mean affinity within a cluster
    min_meanclaff = N.exp( -max_meancldist * max_meancldist / float(var) );

    #---------------------------------------------
    #information of distance within clusters
    if nomatrix:

        #convert fffgraph to scipy sparse matrix
        sparse_mat = createSparceScipyMatrixFromfffGraph( graphfile )
        dist_data = calculate_top_dist_nomatrix( WFV, sparse_mat, top, min_claff )
    #else:

        #mat = aims.read( all_centers_distmat )
        #data = mat.arraydata()[ 0 ][ 0 ]
        #calculate distance data
        #dist_data = calculate_top_dist( WFV, data, top, max_cldist )

    count_divide = 0 #tmp
    singles2b = []
    #review only non-tight clusters
    defcl = len( top )
    finalcl = []
    top2 = []

    for i in range( defcl ):

        maxd = dist_data[ i ]
        divide = False
        if nomatrix:

            if maxd <= min_claff: #maxd in the minimum affinity

                divide = True

        if divide:

           #add top chidren to divide queue
            count_divide += 1 #tmp
            t = top[ i ]
            children = WFV.ch[ t ]
            if ( children[ 0 ] != -1 or children[ 1 ] != -1 ):

                for chi in children:

                    if chi >= Vg:

                        top2.append( chi )
                    else:

                        singles2b.append( chi )
        else:

            finalcl.append( top[ i ] )

    #analisys queue
    queue = []
    #add all trees in queue
    for i in range( len( top2 ) ):

        queue.append( top2[ i ] )

    singles2 = []
    #perform analisys
    while len( queue ) > 0:

        ind = queue[ 0 ]
        queue.remove( ind )
        if nomatrix:

            maxd = calculate_node_dist_nomatrix( WFV,
                                                 sparse_mat,
                                                 ind,
                                                 min_claff )


        children = WFV.ch[ ind ]
        #if maxd > max_cldist and meand > max_meancldist:
        divide = False
        if nomatrix:

            if maxd <= min_claff: #maxd in the minimum affinity

                divide = True

        if divide:

            count_divide += 1 #tmp
            #must divide
            if ( WFV.ch[ Vg ][ 0 ] != -1 and WFV.ch[ Vg ][ 1 ] != -1 ):

                for chi in children:

                    if chi >= Vg:

                        queue.append( chi )
                    else:

                        singles2b.append( chi )
            else:

                singles2.append( ind )
        else:

            if ( WFV.ch[ Vg ][ 0 ] != -1 and WFV.ch[ Vg ][ 1 ] != -1 ):

                finalcl.append( ind )
            else:

                singles2.append( ind )

    WFV.singles2 = singles2 + singles2b
    WFV.finalcl_singles = singles + singles2 + singles2b
    WFV.finalcl_notsingles = finalcl
    WFV.finalcl = finalcl + WFV.finalcl_singles
    #print ('total singles num = ', len( WFV.finalcl_singles ))
    #print ('not single clusters = ', len( WFV.finalcl_notsingles ))
    #print ('total clusters = ', len( WFV.finalcl ))
    #print ('count divide = ', count_divide)

    if add_singles == True:

        create_partition_from_finalcl( WFV, WFV.finalcl )
    else:

        create_partition_from_finalcl( WFV, WFV.finalcl_notsingles )
    return WFV



import numpy as np
from OpenGL import GL
from .VisualizationBaseObject import VisualizationBaseObject, drawable, config
from .Tools.visualizationEnums import VisualizationObject
from .Shaders import Shader
from .Tools import glm
from .BoundingBox import BoundingBox
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

_vs_vs = files('phybers.fibervis') / 'shaders' / 'volume-slice.vs'
_s_fs = files('phybers.fibervis') / 'shaders' / 'slice.fs'

def configTexture(func):
    ''' Decorator for set the texture filter
    '''

    def wrapper(*args, **kwargs):
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_3D, args[0].texture)

        if not args[0].linearInterp:
            GL.glTexParameteri(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
            GL.glTexParameteri(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)

            func(*args, **kwargs)

            GL.glTexParameteri(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR)
            GL.glTexParameteri(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)

        else:
            func(*args, **kwargs)

    return wrapper


class Slice(VisualizationBaseObject):
    def __init__(self, mri, shaderDict):

        super().__init__(mri, shaderDict)
        self.identifier = VisualizationObject.MRISlice

        self.fileName = 'Slice render'

        self.scaleModel = np.matmul(mri.affine, glm.scaleMatrix(mri.mriData.shape))

        self.dims = np.array(mri.mriData.shape, dtype=np.float32)

        # Matrixs
        self.model = mri.translateMat*mri.rotationMat*mri.scaleMat

        self.inverseModel = mri.inverseModel
        self.scaleMat = mri.scaleMat
        self.translateMat = mri.translateMat
        self.rotationMat = mri.rotationMat

        # visualization parameters
        self.linearInterp = True

        self.discardValues = False
        self.threshold = 0.0

        self.slope = 1.0/mri.mriData.max()
        self.bright = 0.0
        self.contrast = 1.0

        self.pos2Axis = None
        self.axis = [0.0, 0.0, 0.0]
        self.calculatePlaneNormal([1.0, 0.0, 0.0], 0.5)

        self.frontIdx = 0

        self.loadStaticUniforms()

        # MRI data already loaded in the gpu
        self.texture = mri.texture
        self.vao = None
        self.vbo = mri.vbo

        self._loadBuffers()

        # Ready to draw
        self.drawable = True

        self.boundingbox = BoundingBox(shaderDict, self, mri.boundingbox.dims, mri.boundingbox.center, self.scaleModel)

        self.max = mri.mriData.max()


    def cleanOpenGL(self):
        print('Cleaning object: ', self)
        GL.glDeleteVertexArrays(1, [self.vao])

        self.clean = True

    def get_center(self):
        return self.boundingbox.get_center() + self.translateMat[:3, 3].flat

    def get_size(self):
        return self.boundingbox.get_size()

    def loadStaticUniforms(self):
        self.shader[0].glUseProgram()

        v1 = np.array([0,1,3,-1, 1,0,1,3, 0,4,5,-1, 4,0,4,5, 0,2,6,-1, 2,0,2,6], dtype=np.int32)
        v2 = np.array([1,3,7,-1, 5,1,3,7, 4,5,7,-1, 6,4,5,7, 2,6,7,-1, 3,2,6,7], dtype=np.int32)

        nSequence = np.array([  0,1,2,3,4,5,6,7, 1,3,0,2,5,7,4,6, 2,0,3,1,6,4,7,5, 3,2,1,0,7,6,5,4,
                                4,0,6,2,5,1,7,3, 5,1,4,0,7,3,6,2, 6,2,7,3,4,0,5,1, 7,6,3,2,5,4,1,0], dtype=np.int32)

        self.vertexPoints = np.array([  1,1,0,  1,1,1,  0,1,0,  0,1,1,
                                        1,0,0,  1,0,1,  0,0,0,  0,0,1], dtype=np.float32).reshape((8,3))

        GL.glUniform1iv(self.shader[0].glGetUniformLocation("v1"), v1.size, v1)
        GL.glUniform1iv(self.shader[0].glGetUniformLocation("v2"), v2.size, v2)
        GL.glUniform1iv(self.shader[0].glGetUniformLocation("nSequence"), nSequence.size, nSequence)
        GL.glUniform3fv(self.shader[0].glGetUniformLocation("vertexPoints"), self.vertexPoints.size, self.vertexPoints)


    def _loadBuffers(self):
        self.vao = GL.glGenVertexArrays(1)
        GL.glBindVertexArray(self.vao)

        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo)

        # Enable attributes
        vertexIdxAttribute = self.shader[self.selectedShader].attributeLocation('vertexIdx')

        # # Connect attributes
        GL.glEnableVertexAttribArray(vertexIdxAttribute)
        GL.glVertexAttribIPointer(vertexIdxAttribute, 1, GL.GL_INT, 0, None)


    def loadUniforms(self):
        '''  This function is called always before drawing. It makes sure that the current bundle will be visualized with the uniforms specified for it.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''


        # Vertex shader
        GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation("M"), 1, GL.GL_TRUE, self.model.getA())
        GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation("S"), 1, GL.GL_TRUE, self.scaleModel.getA())

        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation("frontIdx"), self.frontIdx)
        GL.glUniform4f(self.shader[self.selectedShader].glGetUniformLocation("np"), *self.np, 0.0)
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation("dPlane"), self.dPlane)

        # Fragment shader
        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation("mriTexture"), 0)
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation("slope"), self.slope)
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation('bright'), self.bright)
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation('contrast'), self.contrast)
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation('threshold'), self.threshold if self.discardValues else 0)


    def calculatePlaneNormal(self, axis, pos2Axis_new, newModel=False):
        self.pos2Axis = pos2Axis_new

        newAxis = np.array(axis[:3], dtype=np.float32)
        newAxis /= np.linalg.norm(newAxis)

        if not np.array_equal(newAxis, self.axis) or newModel:
            self.axis = newAxis

            begin = (self.model*self.scaleModel * np.array([0.0, 0.0, 0.0, 1.0], dtype=np.float32).reshape((4,1)))[:3].ravel()
            end = (self.model*self.scaleModel * np.array([*self.axis, 1.0], dtype=np.float32).reshape((4,1)))[:3].ravel()

            axisLengthV = end - begin

            self.np = (axisLengthV / np.linalg.norm(axisLengthV)).getA().ravel()[:3]

            if (self.axis==0.0).sum() == 2:
                # plane eq: dot(vertex, normal) + D = 0
                self.dPlaneBegin = np.dot(begin, self.np)
                self.dPlaneEnd = np.dot(end, self.np)

            else:
                self.dPlaneBegin = np.dot((self.model*self.scaleModel * np.array([*self.vertexPoints[0], 1.0], dtype=np.float32).reshape((4,1)))[:3].ravel(), self.np)
                self.dPlaneEnd = np.dot((self.model*self.scaleModel * np.array([*self.vertexPoints[0], 1.0], dtype=np.float32).reshape((4,1)))[:3].ravel(), self.np)

                for i in range(1,8):
                    D = np.dot((self.model*self.scaleModel * np.array([*self.vertexPoints[i], 1.0], dtype=np.float32).reshape((4,1)))[:3].ravel(), self.np)
                    if D > self.dPlaneEnd:
                        self.dPlaneEnd = D
                    if D < self.dPlaneBegin:
                        self.dPlaneBegin = D
                        self.frontIdx = i

        self.dPlane = (self.dPlaneEnd - self.dPlaneBegin)*self.pos2Axis + self.dPlaneBegin


    def setLinearInterpolation(self, linearInterp):
        self.linearInterp = linearInterp


    def setDiscardValuesWithThreshold(self, discardValues, threshold):
        self.discardValues = discardValues
        self.threshold = threshold


    def setBright(self, bright):
        self.bright = bright


    def setContrast(self, contrast):
        self.contrast = contrast


    @drawable
    @config
    @configTexture
    def draw(self):
        GL.glDrawArraysInstanced(GL.GL_TRIANGLE_FAN, 0, 6, 1) # so we dont have to create another vertex shader

        self.boundingbox.draw()

    @staticmethod
    def createProgram():
        ''' Anonymous function.
        It creates the shader programs for this specific class and returns the handler.
        '''

        vertexGLSL = [str(_vs_vs)]
        fragmentGLSL = [str(_s_fs)]
        return [Shader(vertexGLSL, fragmentGLSL)]


    def rotate(self, center, angle, axis):
        super().rotate(center, angle, axis)
        self.calculatePlaneNormal(self.axis, self.pos2Axis, True)


    def translate(self, vec):
        super().translate(vec)
        self.calculatePlaneNormal(self.axis, self.pos2Axis, True)


    def stackTranslate(self, vec):
        super().stackTranslate(vec)
        self.calculatePlaneNormal(self.axis, self.pos2Axis, True)


    def scale(self, vec, center):
        super().scale(vec, center)
        self.calculatePlaneNormal(self.axis, self.pos2Axis, True)


    def stackScale(self, vec, center):
        super().stackScale(vec, center)
        self.calculatePlaneNormal(self.axis, self.pos2Axis, True)


    def resetModel(self):
        super().resetModel()
        self.calculatePlaneNormal(self.axis, self.pos2Axis, True)



import itertools
import string
import numpy as np


def slope(curve):
    """
    Slope between points of a 2D curve, if a curve has
    n points returns an array with n - 1 points.
    """
    a = curve[:-1]
    b = np.roll(curve, -1, axis=0)[:-1]

    dxdy = b.T - a.T
    return dxdy[1] / dxdy[0]


def string_prefix(data):
    return np.array([string.ascii_lowercase[i] + str(d)
                     for i, d in enumerate(data)])
def string_slopes(slopes):
    return np.array([string.ascii_lowercase[i] + str(slope)
                     for i, slope in enumerate(np.nditer(slopes))])

def get_slopes(curves, sign=True):
    """
    Get multidimensional array of slopes
    given an array of curves.
    """
    if sign:
        return np.array([np.sign(slope(x)) for x in curves])
    else:
        return np.array([slope(x) for x in curves])


# Turn into generator?
def str_prefix_slopes(slopes):
    """
    Returns slopes with a prefix based on it's position, example:
    [3,2,1] -> ['a3','b2','c1']
    """
    return np.fromiter((string.ascii_lowercase[i] + str(value)
                        for i,value in enumerate(slopes)), dtype=(str, len(slopes)))


# Transform curve to slopes with a string prefix.
def curve_to_str_slopes(curve, sign=True):
    if sign:
        return str_prefix_slopes(np.sign(slope(curve)))
    else:
        return str_prefix_slopes(slope(curve))

# General function to transform an array of curves into
# data for minhashing.
# Here curves is just an iterable.
def encode_curves(curves, encoding_function):
    for curve in curves:
        yield encoding_function(curve)

def flatten_lists(clusters_streamlines):
    return list(itertools.chain.from_iterable(clusters_streamlines))

def streamline_len(s):
    # Aproximate arc length.
    return sum(np.linalg.norm(s[:-1] - s[1:], axis=1))

def get_long_clusters(clusters, min_len):
    """
    Return clusters where the centroids has at least min_len length.
    """
    return [cluster for cluster in clusters
            if streamline_len(cluster.centroid) >= min_len]
def get_short_clusters(clusters, max_len):
    """
    Return clusters where the centroids has at most max_len length.
    """
    return [cluster for cluster in clusters
           if streamline_len(cluster.centroid <= max_len)]
def get_clusters_len(clusters, min_len, max_len):
    """
    Clusters where length(centroids) is between min and max len.
    """
    lens = [streamline_len(cluster.centroid) for cluster in clusters]
    return [cluster for i, cluster in enumerate(clusters)
           if lens[i] >= min_len and lens[i] < max_len]



import collections
import numpy as np
from os import path
from dipy.segment.clustering import ClusterMapCentroid
from . import clustering
from ...utils import write_bundle


def clusters_fibers(results: ClusterMapCentroid, final_bundles_dir, infile):
    for i, cluster in enumerate(results.clusters):
        bcust = [infile[j] for j in cluster.indices]
        write_bundle(path.join(final_bundles_dir, f'{i}.bundles'), bcust)

def save_clusters(dataset, clusters, filename):
    """
    Generate output file for clusters, the output file consists of
    the index to which cluster each element in the dataset(streamlines)
    belongs to, an index of -1 indicates that the element isnt assigned anywhere
    """
    output_array = -1*np.ones(len(dataset), np.int32)
    for cluster in clusters:
        output_array[cluster.indices] = cluster.id
    np.savetxt(X=output_array, fname=filename, delimiter='\n', fmt='%d')

def save_clusters_centroids(clusters, filename):
    with open(filename,"w") as file:
        for cluster in clusters:
            file.write(f"{cluster.centroid} \n")

def save_join_clusters_fibers(results, filename):
    with open(filename,"w") as file:
        for clusters_names in results:
            for name in clusters_names:
                text = name + "-" + " ".join(map(str, clusters_names[name].indices)) + "\n"
                file.write(text)

def load_clusters(dataset, filename, not_assigned=False):
    """
    From file where each line is the id of the cluster to which each streamline belongs
    generate a dipy ClusterMapCentroid to use
    """

    cls_dict = collections.defaultdict(list)
    with open(filename, 'r') as f:
        for i,id in enumerate(f):
            id = int(id)
            cls_dict[id].append(i)

    no_clusters = cls_dict.pop(-1)
    clusters = clustering.clusters_to_clustermap(list(cls_dict.values()), dataset, cls_dict.keys())

    if not_assigned:
        return clusters, no_clusters
    else:
        return clusters



import sys
from .segment import fiberseg

if __name__ == "__main__":
    if len(sys.argv) == 6:
        fiber_input = sys.argv[1]
        idsubj = sys.argv[2]
        atlasdir = sys.argv[3]
        atlasInformation = sys.argv[4]
        result_path = sys.argv[5]
        fiberseg(fiber_input, idsubj, atlasdir, atlasInformation, result_path)
    else:
        print(fiberseg.__doc__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import numpy as np
import pandas as pd

#%%
def getBundleSize( bundlefile ):
    """ Get Bundles Size from file.
    """
    # Get center names from bundle file.
    ns = dict()
    exec(compile(open( bundlefile ).read(), bundlefile, 'exec'), ns)
    size = ns[ 'attributes' ][ 'curves_count' ]
    return  size

def fiber_lens(fiber):
    """
    Find the length of a fiber.
    """
    final_len = 0
    for i in range(fiber.shape[0]-1):
        current_lens = np.linalg.norm(fiber[i,:] - fiber[i+1,:])
        final_len+=current_lens
    return final_len


def Stadistics_txt_toxlsx(dirstad):
    """Create Dataframe from statistical measures: sizes, lens, intra_means,intra_min and Ri_dbindex.
    """
    with open(dirstad+"sizes.txt", "r") as file:
        tam = [int(i[:-1]) for i in file]

    with open(dirstad+"lens.txt", "r") as file:
        larg = [round(float(i[:-1]), 2) for i in file]

    with open(dirstad+"mensure_intra_dists.txt", "r") as file:
        intra_max = [round(float(i.split(" ")[1][:-1]), 2) for i in file]

    with open(dirstad+"mensure_intra_means_dists.txt", "r") as file:
        intra_mean = [round(float(i.split(" ")[1][:-1]), 2) for i in file]

    with open(dirstad+"mensure_Ri_db.txt", "r") as file:
        Ri_db = [round(float(i.split(" ")[1][:-1]), 2) for i in file]

    data= {'sizes':tam,'lens':larg,'intra_max':intra_max,'intra_mean':intra_mean,'Ri_db':Ri_db }
    df_data = pd.DataFrame(data, columns = ['sizes','lens','intra_max','intra_mean','Ri_db'])
    df_data.to_excel(dirstad+'stadistics.xlsx')



from .Quaternion import Quaternion
import numpy as np

def null(w=4, h=4):
    return np.matrix([[ 0 for i in range(w)] for j in range(h)], dtype = "float32")


def identity(w=4, h=4):
    return np.matrix([[1 if i==j else 0 for i in range(w)] for j in range(h)], dtype = "float32")


def normalize(v):
    module = np.sqrt(np.multiply(v[0:3],v[0:3]).sum())

    v /= module

    return v


def translateMatrix(v):
    return np.matrix([      [1,     0,      0,      v[0]],
                                            [0,     1,      0,      v[1]],
                                            [0,     0,      1,      v[2]],
                                            [0,     0,      0,      1]], dtype=np.float32)


def scaleMatrix(v):
    if hasattr(v, '__iter__'):
        return np.matrix([      [v[0],  0,              0,              0],
                                                [0,             v[1],   0,              0],
                                                [0,             0,              v[2],   0],
                                                [0,             0,              0,              1]], dtype=np.float32)
    else:
        return np.matrix([      [v,             0,              0,              0],
                                                [0,             v,              0,              0],
                                                [0,             0,              v,              0],
                                                [0,             0,              0,              1]], dtype=np.float32)


def rotateMatrix(alfa, axis):
    return Quaternion.fromAngleAxis(alfa*np.pi/180, axis).rotation4Matrix()


def ortho(left, right, bottom, top, zNear, zFar):
    Result = identity()
    Result[0,0] = 2.0 / (right - left);
    Result[1,1] = 2.0 / (top - bottom);
    Result[2,2] = -2.0 / (zFar - zNear);
    Result[3,0] = -(right + left) / (right - left);
    Result[3,1] = -(top + bottom) / (top - bottom);
    Result[3,2] = -(zFar + zNear) / (zFar - zNear);

    return Result;


def frustum(left, right, bottom, top, nearVal, farVal):
    Result = null()
    Result[0,0] = (2.0 * nearVal) / (right - left)
    Result[1,1] = (2.0 * nearVal) / (top - bottom)
    Result[2,0] = (right + left) / (right - left)
    Result[2,1] = (top + bottom) / (top - bottom)
    Result[2,2] = -(farVal + nearVal) / (farVal - nearVal)
    Result[2,3] = -1.0
    Result[3,2] = -(2.0 * farVal * nearVal) / (farVal - nearVal)

    return Result


def perspective(fovy, aspect, zNear, zFar):
    rad = np.radians(fovy)

    tanHalfFovy = np.tan(rad/2.0)

    Result = null()
    Result[0,0] = 1.0/(aspect*tanHalfFovy)
    Result[1,1] = 1.0/tanHalfFovy
    Result[2,2] = -(zFar + zNear)/(zFar - zNear)
    Result[2,3] = -(2*zFar*zNear)/(zFar - zNear)
    Result[3,2] = -1.0


    return Result


def lookAt(eye, center, up):
    # f = normalize(eye-center)
    # u = normalize(up)
    # s = normalize(np.cross(u, f))
    # u = np.cross(f,s)

    # rotation = identity()

    # rotation[0,0] = s[0];
    # rotation[0,1] = s[1];
    # rotation[0,2] = s[2];

    # rotation[1,0] = u[0];
    # rotation[1,1] = u[1];
    # rotation[1,2] = u[2];

    # rotation[2,0] = f[0];
    # rotation[2,1] = f[1];
    # rotation[2,2] = f[2];

    # rotation[3,0] = -eye[0]
    # rotation[3,1] = -eye[1]
    # rotation[3,2] = -eye[2]

    f = normalize(center-eye)
    u = normalize(up)
    s = normalize(np.cross(f, u))
    u = np.cross(s, f)

    Result = identity()

    Result[0,0] = s[0];
    Result[1,0] = s[1];
    Result[2,0] = s[2];
    Result[0,1] = u[0];
    Result[1,1] = u[1];
    Result[2,1] = u[2];
    Result[0,2] = -f[0];
    Result[1,2] = -f[1];
    Result[2,2] = -f[2];
    Result[3,0] = -np.dot(s, eye)
    Result[3,1] = -np.dot(u, eye)
    Result[3,2] = np.dot(f, eye)

    return Result

    # print(rotation)

    # return rotation.T


def inverseTranspose(m):
    SubFactor00 = m[2,2] * m[3,3] - m[3,2] * m[2,3]
    SubFactor01 = m[2,1] * m[3,3] - m[3,1] * m[2,3]
    SubFactor02 = m[2,1] * m[3,2] - m[3,1] * m[2,2]
    SubFactor03 = m[2,0] * m[3,3] - m[3,0] * m[2,3]
    SubFactor04 = m[2,0] * m[3,2] - m[3,0] * m[2,2]
    SubFactor05 = m[2,0] * m[3,1] - m[3,0] * m[2,1]
    SubFactor06 = m[1,2] * m[3,3] - m[3,2] * m[1,3]
    SubFactor07 = m[1,1] * m[3,3] - m[3,1] * m[1,3]
    SubFactor08 = m[1,1] * m[3,2] - m[3,1] * m[1,2]
    SubFactor09 = m[1,0] * m[3,3] - m[3,0] * m[1,3]
    SubFactor10 = m[1,0] * m[3,2] - m[3,0] * m[1,2]
    SubFactor11 = m[1,1] * m[3,3] - m[3,1] * m[1,3]
    SubFactor12 = m[1,0] * m[3,1] - m[3,0] * m[1,1]
    SubFactor13 = m[1,2] * m[2,3] - m[2,2] * m[1,3]
    SubFactor14 = m[1,1] * m[2,3] - m[2,1] * m[1,3]
    SubFactor15 = m[1,1] * m[2,2] - m[2,1] * m[1,2]
    SubFactor16 = m[1,0] * m[2,3] - m[2,0] * m[1,3]
    SubFactor17 = m[1,0] * m[2,2] - m[2,0] * m[1,2]
    SubFactor18 = m[1,0] * m[2,1] - m[2,0] * m[1,1]

    Inverse = null()
    Inverse[0,0] = + (m[1,1] * SubFactor00 - m[1,2] * SubFactor01 + m[1,3] * SubFactor02)
    Inverse[0,1] = - (m[1,0] * SubFactor00 - m[1,2] * SubFactor03 + m[1,3] * SubFactor04)
    Inverse[0,2] = + (m[1,0] * SubFactor01 - m[1,1] * SubFactor03 + m[1,3] * SubFactor05)
    Inverse[0,3] = - (m[1,0] * SubFactor02 - m[1,1] * SubFactor04 + m[1,2] * SubFactor05)

    Inverse[1,0] = - (m[0,1] * SubFactor00 - m[0,2] * SubFactor01 + m[0,3] * SubFactor02)
    Inverse[1,1] = + (m[0,0] * SubFactor00 - m[0,2] * SubFactor03 + m[0,3] * SubFactor04)
    Inverse[1,2] = - (m[0,0] * SubFactor01 - m[0,1] * SubFactor03 + m[0,3] * SubFactor05)
    Inverse[1,3] = + (m[0,0] * SubFactor02 - m[0,1] * SubFactor04 + m[0,2] * SubFactor05)

    Inverse[2,0] = + (m[0,1] * SubFactor06 - m[0,2] * SubFactor07 + m[0,3] * SubFactor08)
    Inverse[2,1] = - (m[0,0] * SubFactor06 - m[0,2] * SubFactor09 + m[0,3] * SubFactor10)
    Inverse[2,2] = + (m[0,0] * SubFactor11 - m[0,1] * SubFactor09 + m[0,3] * SubFactor12)
    Inverse[2,3] = - (m[0,0] * SubFactor08 - m[0,1] * SubFactor10 + m[0,2] * SubFactor12)

    Inverse[3,0] = - (m[0,1] * SubFactor13 - m[0,2] * SubFactor14 + m[0,3] * SubFactor15)
    Inverse[3,1] = + (m[0,0] * SubFactor13 - m[0,2] * SubFactor16 + m[0,3] * SubFactor17)
    Inverse[3,2] = - (m[0,0] * SubFactor14 - m[0,1] * SubFactor16 + m[0,3] * SubFactor18)
    Inverse[3,3] = + (m[0,0] * SubFactor15 - m[0,1] * SubFactor17 + m[0,2] * SubFactor18)

    Determinant = m[0,0] * Inverse[0,0] + m[0,1] * Inverse[0,1] + m[0,2] * Inverse[0,2] + m[0,3] * Inverse[0,3]

    Inverse /= Determinant

    return Inverse



import shlex
import sys
import time
from ..utilities import inter2bundles

#%% 7. Intersección para ello se usa la función que se muestra a continuación (inter2bundles) realizada en python y esta ejecuta un programa de c que calcula
# que se llama "fiberDistanceMax2bun.c"

# Directorios de los dos set de fibras a comprarar

dir_fib1 = sys.argv[1]
dir_fib2 = sys.argv[2]
outdir= sys.argv[3] # este dir es para guardar cálculos como la matriz

d_th = float(sys.argv[4]) # umbral de intersección en milimetros y lo define el usuario. default: 15

pinter = inter2bundles (dir_fib1, dir_fib2, outdir, d_th)
print ("Intersection in the set of fibers1 [%]: ", pinter[0])
print ("Intersection in the set of fibers2 [%]:: ", pinter[1])

t0=time.time()

print ("Demora: ", time.time()-t0 ," [s]" )



import math
import operator
import numpy as np
from random import randint
from dipy.tracking.streamline import length as streamline_length


def recalc_centroid(indices, refdata):
    fibers = {}
    for i in indices:
        f = refdata[i]
        f_lenght = streamline_length(f)
        fibers[i] = f_lenght
    sorted_items = sorted(fibers.items(), key=operator.itemgetter(1))
    sorted_fibers = []
    for item in sorted_items:
        sorted_fibers.append(item[0])
    index_60length = math.ceil(len(sorted_fibers)*0.6) - 1
    index_80length = math.ceil(len(sorted_fibers)*0.8) - 1
    longest_fibers = []
    while index_60length <= index_80length:
        longest_fibers.append(sorted_fibers[index_60length])
        index_60length+=1
    number_fibers_10percent = math.ceil(len(longest_fibers)*0.1)
    fibers_10percent = []
    for i in range(number_fibers_10percent):
        rand_index = randint(1, number_fibers_10percent)
        fibers_10percent.append(longest_fibers[rand_index])

    fibers_data = []
    for index in fibers_10percent:
        fiber = refdata[index]
        fibers_data.append(fiber)
    matrix = matrix_dist(streamlines = np.asarray(fibers_data))

    assert matrix is not None

    min_sum = 0
    for i in range(len(matrix)):
        sum = 0
        for j in range(len(matrix[0])):
            sum += matrix[i][j]
        if (sum < min_sum):
            min_sum = sum

    centroid = fibers_data[len(matrix) - 1]
    return centroid

def matrix_dist(streamlines, get_max=True, get_mean=False):
    """
    Compute distance between all streamlines, can return both max/min distance matrices.
    """
    x = streamlines
    # Output of this computation is a (num_streamlines, num_streamlines, num_points) array.
    # First stack both direct and reversed in an array -> (2, num_streamlines, num_points, dim)
    # using [:,None] create new axis to broadcast computation and perform difference between
    # every pair of streamlines(both direct and reverse), then perform x^2 + y^2 + z^2.
    distances = np.sum(np.square(np.stack((x,x[:,::-1]))[:,None]-x[:,None]), axis=4)

    if get_max and get_mean:
        max_distances = np.sqrt(distances.max(axis=3).min(axis=0))
        mean_distances = np.sqrt(distances.mean(axis=3).min(axis=0))
        return max_distances, mean_distances
    elif get_max:
        return np.sqrt(distances.max(axis=3).min(axis=0))
    elif get_mean:
        return  np.sqrt(distances.mean(axis=3).min(axis=0))
    else:
        print('error, should return atleast one matrix')



# Works with Atlas that contains 254 or less bundles... For more bundles fiberValidator should be uint16, not uint8

import numpy as np
from OpenGL import GL
from .SegmentationHandler import SegmentationHandler
from ..Tools.utilities import findIntegersMultiplierFor
from ..Tools.visualizationEnums import SegmentationTypes
from ..BoundingBox import BoundingBox
from ..VisualizationBaseObject import VisualizationBaseObject, propagateToChildren, drawable, config
from ..Tools.visualizationEnums import VisualizationObject
from ..Shaders import Shader
from ...FiberVis_core import reSampleBundle, AtlasBasedSegmentation, AtlasBasedSegmentationExportbundlesdata
from ..Bundle import Bundle
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

_abs_vs = files('phybers.fibervis') / 'shaders' / 'atlasbasedsegmentation.vs'
_sfs_fs = files('phybers.fibervis') / 'shaders' / 'standardFragmentShader.fs'
_abs_gs = files('phybers.fibervis') / 'shaders' / 'atlasbasedsegmentation.gs'


class Atlas(VisualizationBaseObject):
    def __init__(self, shaderDict, parent):
        super().__init__(parent, shaderDict[Bundle])
        self.identifier = VisualizationObject.Atlas

        self.fileName = 'Atlas'

        self.bundlesNames = []
        self.bundlesNFiber = []
        self.points = []
        self.colorTable = np.array([0.5,0.5,0.5,2.0], dtype=np.float32).reshape((1,4))

        self.bundlesData = []

        self.thresholdNames = []
        self.thresholds = None
        self.thresholdNFiber = []

        self.validBundles = []
        self.validBundlesIndex = []
        self.missingBundle = []
        self.missingThreshold = []

        # Ready to draw
        self.drawable = True

        self.boundingbox = BoundingBox(shaderDict, self, [0,0,0], [0,0,0])

    def get_center(self):
        return self.boundingbox.get_center() + self.translateMat[:3, 3].flat

    def get_size(self):
        return self.boundingbox.get_size()


    def addBundle(self, sPathList):

        newItems = len(self.children)

        for sPath in sPathList:
            if sPath in [k.path for k in self.children]:
                continue

            bundleHolder = Bundle(sPath, self.shader, self)
            bundleHolder.boundingbox.drawable = False

            self.children.append(bundleHolder)

        print("Loading ready:\nFiles added to Atlas:\t", [k.fileName for k in self.children[newItems:]])

        self.createSegmentationData()


    def removeBundle(self, bundleList):

        tmpFileNameList = [i.fileName for i in self.children]
        removedItems = sorted([tmpFileNameList.index(i) for i in bundleList], reverse=True)

        for rmIndex in removedItems:
            del self.children[rmIndex]

        # self.points = np.empty(self.points.size-removedPoints, dtype=np.float32)
        # np.concatenate([i.points for i in self.children], out=self.points)
        # # point bundles.points to atlas.points?

        # self.colorTable = np.empty(len(self.bundlesNames)*4, dtype=np.float32)
        # np.concatenate([i.colorTable for i in self.children], out=self.colorTable)                    # Color for each Bundle

        # # We set the boundingbox parameters
        # x = self.points[0::3]
        # y = self.points[1::3]
        # z = self.points[2::3]

        # xmin, xmax = x.min(),x.max()
        # ymin, ymax = y.min(),y.max()
        # zmin, zmax = z.min(),z.max()

        # dims = np.array([xmax-xmin, ymax-ymin, zmax-zmin], dtype=np.float32)
        # center = np.array([xmin, ymin, zmin], dtype=np.float32)

        # self.boundingbox = BoundingBox(shaderDict, self, dims, center)

        print("Files removed from Atlas:\t", bundleList)

        self.createSegmentationData()


    def loadThresholdFile(self, thrFile):
        with open(thrFile) as file:
            lines = [line.rstrip('\n') for line in file]

            self.thresholdNames = []
            self.thresholds = np.empty(len([validLine for validLine in lines if len(validLine) != 0]), dtype=np.uint8)
            self.thresholdNFiber = []

            for l in lines:
                if len(l) == 0:
                    continue

                name, threshold, nFiber = l.split()

                idx = len(self.thresholdNames)

                self.thresholdNames.append(name)
                self.thresholds[idx] = int(threshold)
                self.thresholdNFiber.append(int(nFiber))

        self.createSegmentationData()


    def calculateValidBundles(self):
        a = set(self.thresholdNames)
        b = self.bundlesNames

        onlyInA = a.difference(b)
        d = a.symmetric_difference(b)
        onlyInB = d.difference(a)

        self.validBundles = list(a.intersection(b))
        self.validBundlesIndex = sorted([self.bundlesNames.index(i) for i in self.validBundles])
        self.missingBundle = list(onlyInA)
        self.missingThreshold = list(onlyInB)


    def getValidBundles(self):
        validBundlesPacked = []

        for validBundleName in self.validBundles:
            validBundleFiberN = self.bundlesNFiber[self.bundlesNames.index(validBundleName)]
            validBundleHie = self.colorTable[self.bundlesNames.index(validBundleName)]
            assert isinstance(self.thresholds, np.ndarray)
            validBundleThreshold = self.thresholds[self.thresholdNames.index(validBundleName)]

            validBundlesPacked.append((validBundleName, validBundleFiberN, validBundleHie, validBundleThreshold))

        return validBundlesPacked

    def getInvalidBundles(self):
        invalidBundlesPacked = []

        for invalidBundleName in self.missingBundle:

            validBundleFiberN = 0
            validBundleHie = 0
            assert isinstance(self.thresholds, np.ndarray)
            validBundleThreshold = self.thresholds[self.thresholdNames.index(invalidBundleName)]

            invalidBundlesPacked.append((invalidBundleName, validBundleFiberN, validBundleHie, validBundleThreshold))

        for invalidBundleName in self.missingThreshold:

            validBundleFiberN = self.bundlesNFiber[self.bundlesNames.index(invalidBundleName)]
            validBundleHie = self.colorTable[self.bundlesNames.index(invalidBundleName)]
            validBundleThreshold = 0

            invalidBundlesPacked.append((invalidBundleName, validBundleFiberN, validBundleHie, validBundleThreshold))

        return invalidBundlesPacked




    def createSegmentationData(self):
        self.bundlesNames = [i for k in self.children for i in k.bundlesName]
        self.bundlesNFiber = [i for k in self.children for i in k.bundlesInterval[1:]-k.bundlesInterval[:-1]]
        self.points = [k.points[k.fiberSizes[:k.bundlesInterval[i]].sum()*3 : k.fiberSizes[:k.bundlesInterval[i+1]].sum()*3] for k in self.children for i in range(len(k.bundlesName))]

        self.colorTable = np.empty((len(self.bundlesNames)+1, 4), dtype=np.float32)
        np.concatenate([i.colorTable for i in self.children] + [np.array([0.5,0.5,0.5,2.0], dtype=np.float32).reshape((1,4))], out=self.colorTable)

        self.calculateValidBundles()

        self.fiber2Bundle = []

        totalSize = 0

        for i in  self.points:
            totalSize += i.size

        fiber2BundleHolder = np.empty(totalSize//3, dtype=np.uint8)

        offset = 0
        for i in range(len(self.points)):
            self.fiber2Bundle.append(fiber2BundleHolder[offset:offset+self.points[i].size//63])
            self.fiber2Bundle[-1][:] = i
            offset += self.points[i].size//3

        if not len(self.validBundlesIndex):
            return

        self.validPoints = np.concatenate([self.points[i] for i in self.validBundlesIndex])
        self.validFiber2Bundle = np.concatenate([self.fiber2Bundle[i] for i in self.validBundlesIndex])

        self.validThreshold = np.zeros(len(self.bundlesNames), dtype=np.uint8)

        for i in self.validBundlesIndex:
            assert isinstance(self.thresholds, np.ndarray)
            self.validThreshold[i] = self.thresholds[self.thresholdNames.index(self.bundlesNames[i])]

        self.calculateBoundingBox()


    def calculateBoundingBox(self):

        # We set the boundingbox parameters
        x = self.children[0].points[0::3]
        y = self.children[0].points[1::3]
        z = self.children[0].points[2::3]

        xmin, xmax = x.min(),x.max()
        ymin, ymax = y.min(),y.max()
        zmin, zmax = z.min(),z.max()

        for bundle in self.children[1:]:
            x = bundle.points[0::3]
            y = bundle.points[1::3]
            z = bundle.points[2::3]

            xmin_i, xmax_i = x.min(),x.max()
            ymin_i, ymax_i = y.min(),y.max()
            zmin_i, zmax_i = z.min(),z.max()

            if xmin_i < xmin:
                xmin = xmin_i
            if xmax_i > xmax:
                xmax = xmax_i

            if ymin_i < ymin:
                ymin = ymin_i
            if ymax_i > ymax:
                ymax = ymax_i

            if zmin_i < zmin:
                zmin = zmin_i
            if zmax_i > zmax:
                zmax = zmax_i

        dims = np.array([xmax-xmin, ymax-ymin, zmax-zmin], dtype=np.float32)
        center = np.array([xmin, ymin, zmin], dtype=np.float32)

        self.boundingbox.updateBBModel(dims, center)


    def loadAndApplyMatrix(self, matrixFile):
        extension = matrixFile.split('.')[-1]

        # numpy matrix file
        if extension == 'npy':
            transform = Bundle.loadNumpyMatrix(matrixFile)

        # brain visa (trm) matrix file
        elif extension == 'trm':
            transform = Bundle.loadTrmMatrix(matrixFile)

        # Unsopported file
        else:
            raise TypeError('Unsupported matrix file.', extension)

        for child in self.children:
            if isinstance(child, Bundle):
                child.applyMatrix(transform)

        self.createSegmentationData()



class AtlasBasedParallelSegmentation(SegmentationHandler):
    fiberLengthTimes3 = 21*3

    def __init__(self, bundle, shaderDict):
        super().__init__(bundle, shaderDict)

        self.segmentationIdentifier = SegmentationTypes.AtlasBased

        self.fileName = 'Atlas based segmentation' # temporal

        self.atlas = Atlas(shaderDict, self)

        self.alpha = 0.8

        self.colorTableTexture = None
        self._loadColorTexture()

        self.configFiberValidator()

        # Creates VBOs & an EBO, then populates them with the point, normal and color data. The elements data goes into the EBO
        self._loadBuffers()
        self.buildVertex2Fiber()
        self.vboAndLinkVertex2Fiber()

        # Check if bundle has fixed point size per fiber (21 vertex per fiber)
        if (self.fiberSizes != 21).sum():
            print('Resampling...')
            self.fixedPoints = np.empty(self.curvescount*self.fiberLengthTimes3, dtype=np.float32)
            reSampleBundle(
                self.point,
                self.fiberSize,
                self.curvescount,
                self.fixedPoint,
                self.fiberLengthTimes3//3)

            # cfuncs.reSampleBundle(
            #   self.points.ctypes.data,
            #   self.fiberSizes.ctypes.data,
            #   self.curvescount,
            #   self.fixedPoints.ctypes.data,
            #   self.fiberLengthTimes3//3)
            print('Done.')
        else:
            self.fixedPoints = self.points

        self.segmentedBundles = []
        self.validBundles = []

        self.boundingbox = BoundingBox(shaderDict, self, bundle.boundingbox.dims, bundle.boundingbox.center)


    def cleanOpenGL(self):
        super().cleanOpenGL()

        GL.glDeleteTextures([self.colorTableTexture])



    def defaultFiberValidator(self):
        return np.zeros(self.validFiberTexDims[1]*self.validFiberTexDims[0], dtype=np.uint8)


    def segmentMethod(self):

        self.fiberValidator[:self.curvescount] = len(self.atlas.bundlesNames)
        self.validBundles = self.atlas.validBundles

        self.updateColorTexture()

        AtlasBasedSegmentation(
            self.atlas.validPoints,
            self.atlas.validPoints.size,
            self.fixedPoints,
            self.fixedPoints.size,
            self.fiberLengthTimes3,
            self.atlas.validThreshold,
            self.atlas.validFiber2Bundle,
            self.fiberValidator)

        # cfuncs.AtlasBasedSegmentation(
        #       self.atlas.validPoints.ctypes.data,                                     # ctypes.c_void_p,                      # float *atlas_data                             vector with all the 3d points for the atlas
        #       self.atlas.validPoints.size,                                            # ctypes.c_int,                         # unsigned int atlas_data_size  size of the vector with all the points for the atlas
        #       self.fixedPoints.ctypes.data,                                           # ctypes.c_void_p,                      # float *subject_data                   vector with all the 3d points for the subject
        #       self.fixedPoints.size,                                                          # ctypes.c_int,                         # unsigned int subject_data_size        size of the vector with all the points for the subject
        #       self.fiberLengthTimes3,                                                         # ctypes.c_int,                         # unsigned short int ndata_fiber        number of points per fiber (*3) so 21 points = 63
        #       self.atlas.validThreshold.ctypes.data,                          # ctypes.c_void_p,                      # unsigned char *thresholds             vector with the thresholds for each fascicle on the atlas
        #       self.atlas.validFiber2Bundle.ctypes.data,                       # ctypes.c_void_p,                      # unsigned int *bundle_of_fiber vector of atlas_points_size with id for the fascicle that correspondence
        #       self.fiberValidator.ctypes.data)                                        # ctypes.c_void_p)                      # unsigned char *assignment             size nfibers_subject. And all data set to 254 - result vector

        self.segmentedBundles = [(self.atlas.bundlesNames[self.atlas.validBundlesIndex[i]], (self.fiberValidator[:self.curvescount] == i).sum(), i)
                                    for i in range(len(self.atlas.validBundles))
                                    if (self.fiberValidator[:self.curvescount] == i).sum() != 0]

        # np.save("D:/Desktop/labels", self.fiberValidator[:self.curvescount])


    def updateAtlas(self, newBundles=None, deleteBundles=None, thresholdFile=None, applyMatrixFile=None):
        if newBundles:
            self.atlas.addBundle(newBundles)

        if deleteBundles:
            self.atlas.removeBundle(deleteBundles)

        if thresholdFile:
            self.atlas.loadThresholdFile(thresholdFile)

        if applyMatrixFile:
            self.atlas.loadAndApplyMatrix(applyMatrixFile)


    def _loadColorTexture(self):
        ''' It generates a texture (if not already). Then makes the texture0 the active one and loads the color table as a 1D texture.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''
        maxTexDim = GL.glGetIntegerv(GL.GL_MAX_TEXTURE_SIZE)

        if maxTexDim*maxTexDim < len(self.atlas.bundlesNames)+1:
            raise ValueError('To many bundles inside the atlas... Color texture not big enought.')
        self.validBundleColorTexDims = findIntegersMultiplierFor(len(self.atlas.bundlesNames)+1, maxTexDim)

        if self.colorTableTexture == None:
            self.colorTableTexture = GL.glGenTextures(1)

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.colorTableTexture)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_BORDER)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_BORDER)

        bgColor = np.array([1.0, 1.0, 1.0, 1.0], dtype=np.float32)
        GL.glTexParameterfv(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_BORDER_COLOR, bgColor)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)

        # GL.glTexImage1D(GL.GL_TEXTURE_1D, 0, GL.GL_RGBA, 1, 0, GL.GL_RGBA, GL.GL_FLOAT, self.atlas.colorTable.flatten())
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, *self.validBundleColorTexDims, 0, GL.GL_RGBA, GL.GL_FLOAT, self.atlas.colorTable.flatten())


    def updateColorTexture(self):
        maxTexDim = GL.glGetIntegerv(GL.GL_MAX_TEXTURE_SIZE)

        if maxTexDim*maxTexDim < len(self.atlas.bundlesNames)+1:
            raise ValueError('To many bundles inside the atlas... Color texture not big enought.')
        self.validBundleColorTexDims = findIntegersMultiplierFor(len(self.atlas.bundlesNames)+1, maxTexDim)

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.colorTableTexture)

        # GL.glTexImage1D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, len(self.atlas.bundlesNames)+1, 0, GL.GL_RGBA, GL.GL_FLOAT, self.atlas.colorTable.flatten())
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, *self.validBundleColorTexDims, 0, GL.GL_RGBA, GL.GL_FLOAT, self.atlas.colorTable.flatten())


    def _loadBuffers(self):
        ''' It generates a VAO, VBO and EBO.
        It populates them with the vertex, normal and colors (the VBO), also the elements (the EBO).

        It finishes when the buffers are linked with attributes on the vertex shader.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        # Reference for vbos, ebos and attribute ptrs
        self.vao = GL.glGenVertexArrays(1)
        GL.glBindVertexArray(self.vao)

        # self.ebo = GL.glGenBuffers(1)

        # VBO positions
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[0])

        positionAttribute =     self.shader[0].attributeLocation('vertexPos')
        GL.glEnableVertexAttribArray(positionAttribute)
        GL.glVertexAttribPointer(positionAttribute,3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)

        # VBO normals
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[1])

        normalAttribute =       self.shader[0].attributeLocation('vertexNor')
        GL.glEnableVertexAttribArray(normalAttribute)
        GL.glVertexAttribPointer(normalAttribute,       3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)

        # EBO
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.ebo)



    def exportBundleFile(self, outfile):
        if len(self.segmentedBundles) == 0:
            return

        bundlesSelected = np.array([i[2] for i in self.segmentedBundles], dtype=np.uint8)

        AtlasBasedSegmentationExportbundlesdata(
                    outfile+'data',
                    len(self.segmentedBundles),
                    bundlesSelected,
                    self.fiberSizes,
                    self.points,
                    self.fiberSizes.size,
                    self.fiberValidator)

        # cfuncs.AtlasBasedSegmentationExportbundlesdata(
        #       (outfile+'data').encode('utf-8'),
        #       len(self.segmentedBundles),
        #       bundlesSelected.ctypes.data,
        #       self.fiberSizes.ctypes.data,
        #       self.points.ctypes.data,
        #       self.fiberSizes.size,
        #       self.fiberValidator.ctypes.data)

        # wrtie minf file
        minf = """attributes = {\n    'binary' : 1,\n    'bundles' : %s,\n    'byte_order' : 'DCBA',\n    'curves_count' : %s,\n    'data_file_name' : '*.bundlesdata',\n    'format' : 'bundles_1.0',\n    'space_dimension' : 3\n  }\n"""

        bundlesstr = []
        offset = 0

        for name, bundleSize, notUsed in self.segmentedBundles:
            bundlesstr.append(name)
            bundlesstr.append(offset)
            offset += bundleSize

        bundlesstr = str(bundlesstr)
        bundlesstr = bundlesstr[0] + ' ' + bundlesstr[1:-1] + ' ' + bundlesstr[-1]

        with open(outfile, 'w') as file:
            file.write( minf % ( bundlesstr, offset ) )


    def loadUniforms(self):
        super().loadUniforms()

        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('notAssigned'), len(self.validBundles))

        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('colorTable'), 0)
        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('fiberValidator'), 1)


    @propagateToChildren
    @drawable
    @config
    def draw(self):
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.colorTableTexture)

        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.validFiberTexture)

        # Collided fiber
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation("alpha"), 5.0)
        GL.glDrawElements(GL.GL_LINE_STRIP, self.eboSize, GL.GL_UNSIGNED_INT, None)

        # Transparent fiber
        if self.alpha >= 0.01:
            GL.glEnable(GL.GL_BLEND)
            GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
            GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation("alpha"), self.alpha)
            GL.glDrawElements(GL.GL_LINE_STRIP, self.eboSize, GL.GL_UNSIGNED_INT, None)
            GL.glDisable(GL.GL_BLEND)

        self.boundingbox.draw()

    @staticmethod
    def createProgram():
        ''' Anonymous function.
        Default shader program for atlas segmentation.
        It creates the shader programs for this specific class and returns the handler.
        '''

        vertexGLSL = [str(_abs_vs)]
        fragmentGLSL = [str(_sfs_fs)]
        geometryGLSL = [str(_abs_gs)]
        return [Shader(vertexGLSL, fragmentGLSL, geometryGLSL)]



from .VisualizationBaseObject import VisualizationBaseObject, drawable, config
from .Shaders import Shader
from .Tools import glm
from OpenGL import GL
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files
import numpy as np

_bbox_vs = files('phybers.fibervis') / 'shaders' / 'boundingbox.vs'
_sfs_fs = files('phybers.fibervis') / 'shaders' / 'standardFragmentShader.fs'

class BoundingBox(VisualizationBaseObject):
    def __init__(self, shaderDict, parent, dims, center, bbM=None):
        super().__init__(parent, shaderDict)
        if bbM is None:
            self.bbModel = glm.translateMatrix(center)*glm.scaleMatrix(dims)
        else:
            self.bbModel = bbM

        self.dims = np.array(dims, dtype=np.float32)
        self.center = np.array(center, dtype=np.float32)

        self._loadBuffers()

        # Ready to draw
        self.drawable = True


    def cleanOpenGL(self):
        print('Cleaning object: ', self)
        GL.glDeleteVertexArrays(1, [self.vao])

        GL.glDeleteBuffers(1, [self.vbo])
        GL.glDeleteBuffers(1, [self.ebo])

        self.clean = True

    def get_center(self):
        return self.center + self.dims / 2

    def get_size(self):
        return np.max(self.dims)

    def _loadBuffers(self):
        self.vao = GL.glGenVertexArrays(1)
        GL.glBindVertexArray(self.vao)

        self.vbo = GL.glGenBuffers(1)
        self.ebo = GL.glGenBuffers(1)

        boundingbox = np.array([1,      1,      0,
                                1,      1,      1,
                                0,      1,      1,
                                0,      1,      0,
                                1,      0,      0,
                                1,      0,      1,
                                0,      0,      0,
                                0,      0,      1], dtype=np.float32)

        boundingElements = np.array([   0, 1, 1, 2, 2, 3, 3, 6, 0, 3, 0, 4,
                                        1, 5, 2, 7, 4, 6, 4, 5, 5, 7, 6, 7], dtype=np.uint32)

        # VBO
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo)
        GL.glBufferData(GL.GL_ARRAY_BUFFER, boundingbox.nbytes, boundingbox, GL.GL_STATIC_DRAW)

        # EBO
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.ebo)
        GL.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, boundingElements.nbytes, boundingElements, GL.GL_STATIC_DRAW)

        # Enable attributes
        positionAttribute =     self.shader[0].attributeLocation('vertexPos')

        # # Connect attributes
        GL.glEnableVertexAttribArray(positionAttribute)
        GL.glVertexAttribPointer(positionAttribute, 3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)


    def updateBBModel(self, dims, center):
        self.bbModel = glm.translateMatrix(center)*glm.scaleMatrix(dims)

        self.dims = np.array(dims, dtype=np.float32)
        self.center = np.array(center, dtype=np.float32)


    def loadUniforms(self):
        GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation("bbM"), 1, GL.GL_TRUE, self.bbModel.getA())
        GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation("M"), 1, GL.GL_TRUE, self.parent.model.getA())


    @drawable
    @config
    def draw(self):
        GL.glDrawElements(GL.GL_LINES, 24, GL.GL_UNSIGNED_INT, None)

    @staticmethod
    def createProgram():
        ''' Anonymous function.
        It creates the shader programs for this specific class and returns the handler.
        '''

        vertexGLSL = [str(_bbox_vs)]
        fragmentGLSL = [str(_sfs_fs)]
        return [Shader(vertexGLSL, fragmentGLSL)]



''' to-do
        configOpengl
        exportbundlefile
'''

import math
import numpy as np
from OpenGL import GL
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files
from ..VisualizationBaseObject import VisualizationBaseObject, propagateToChildren, drawable, config
from ..Tools.visualizationEnums import VisualizationObject
from ..Shaders import Shader
from ..Bundle import Bundle
from ...FiberVis_core import ROISegmentationExportBundlesdata
from ..Tools.performance import timeit

_s_vs = files('phybers.fibervis') / 'shaders' / 'segmentation.vs'
_sfs_vs = files('phybers.fibervis') / 'shaders' / 'standardFragmentShader.fs'
_s_gs = files('phybers.fibervis') / 'shaders' / 'segmentation.gs'

class SegmentationHandler(VisualizationBaseObject):
    '''
    '''

    def __init__(self, bundle, shaderDict):
        '''
        '''
        super().__init__(bundle, shaderDict)
        self.identifier = VisualizationObject.Segmentation
        self.segmentationIdentifier = None

        if not isinstance(bundle, Bundle):
            raise TypeError('Not a bundle.')

        # self.parent = bundle
        # self.shader = shaderProgram[type(self)]

        # OpenGL buffer references
        self.vao = 0
        self.vbo = bundle.vbo
        self.ebo = bundle.ebo

        # Atlas color table, loads when the atlas is loaded
        self.colorTableTexture = bundle.colorTableTexture
        self.validBundleColorTexDims = bundle.validBundleColorTexDims

        # Matrixs
        self.model = bundle.model

        self.inverseModel = bundle.inverseModel
        self.scaleMat = bundle.scaleMat
        self.translateMat = bundle.translateMat
        self.rotationMat = bundle.rotationMat

        self.points = bundle.points
        # self.elements = np.copy(bundle.elements)
        self.eboSize = bundle.elements.size
        self.fiberSizes = bundle.fiberSizes

        self.bundlesName = bundle.bundlesName
        self.curvescount = bundle.curvescount
        self.bundlesStart = np.append(bundle.bundlesStart, self.curvescount).astype(np.int32)

        # Flags
        self.drawable = False


    def cleanOpenGL(self):
        print('Cleaning object: ', self)
        GL.glDeleteVertexArrays(1, [self.vao])

        GL.glDeleteBuffers(1, [self.vboV2F])

        GL.glDeleteTextures([self.validFiberTexture])

        self.clean = True


    def configOpenGL(self):
        pass


    # @averageTimeit
    @timeit
    def segmentSubject(self):
        ''' This function has to be implemented by all segmentation subclass.

        It will create an ebo or load the data that segments the bundle using an atlas.

        It also has to set the colors for the segmentation, by calling self._loadColorTexture()
        '''
        self.segmentMethod()

        self.updateFiberValidator()


    def segmentMethod(self):
        ''' This function has to be implemented by all segmentation subclass.

        It will create an ebo or load the data that segments the bundle using an atlas.

        It also has to set the colors for the segmentation, by calling self._loadColorTexture()
        '''
        raise TypeError('Segmentation method is not implemented in class: ', type(self))


    def updateFiberValidator(self):
        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.validFiberTexture)

        # GL.glTexSubImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_R8UI, *self.validFiberTexDims, 0, GL.GL_RED_INTEGER, GL.GL_UNSIGNED_BYTE, self.fiberValidator)
        GL.glTexSubImage2D(GL.GL_TEXTURE_2D, 0, 0, 0, *self.validFiberTexDims, GL.GL_RED_INTEGER, GL.GL_UNSIGNED_BYTE, self.fiberValidator)


    def _loadBuffers(self):
        ''' It generates a VAO, VBO and EBO.
        It populates them with the vertex, normal and colors (the VBO), also the elements (the EBO).

        It finishes when the buffers are linked with attributes on the vertex shader.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        # Reference for vbos, ebos and attribute ptrs
        self.vao = GL.glGenVertexArrays(1)
        GL.glBindVertexArray(self.vao)

        # VBO position
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[0])

        positionAttribute =     self.shader[0].attributeLocation('vertexPos')
        GL.glEnableVertexAttribArray(positionAttribute)
        GL.glVertexAttribPointer(positionAttribute,3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)

        # VBO normals
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[1])

        normalAttribute =       self.shader[0].attributeLocation('vertexNor')
        GL.glEnableVertexAttribArray(normalAttribute)
        GL.glVertexAttribPointer(normalAttribute,       3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)

        # VBO colors
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[2])

        colorAttribute =        self.shader[0].attributeLocation('vertexCol')
        GL.glEnableVertexAttribArray(colorAttribute)
        GL.glVertexAttribPointer(colorAttribute,        1, GL.GL_INT, GL.GL_FALSE, 0, None)

        # EBO
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.ebo)


    def loadUniforms(self):
        '''This function is called always before drawing. It makes sure that the current class will be visualized with the uniforms specified for it.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation("M"), 1, GL.GL_TRUE, self.model.getA())

        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('texture1DMax'), self.validFiberTexDims[0])
        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('bundleTex1DMax'), self.validBundleColorTexDims[0])


    def buildVertex2Fiber(self):
        self.vertex2Fiber = np.empty(self.parent.points.size//3, dtype=np.int32)

        offset = 0
        for i in range(len(self.parent.fiberSizes)):
            self.vertex2Fiber[offset:offset+self.parent.fiberSizes[i]] = i
            offset += self.parent.fiberSizes[i]


    def vboAndLinkVertex2Fiber(self):
        GL.glBindVertexArray(self.vao)

        self.vboV2F = GL.glGenBuffers(1)

        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vboV2F)
        GL.glBufferData(GL.GL_ARRAY_BUFFER, self.vertex2Fiber.nbytes, self.vertex2Fiber, GL.GL_STATIC_DRAW)

        # Enable attributes
        v2FAttribute = self.shader[0].attributeLocation('vertex2Fib')

        # Connect attributes
        GL.glEnableVertexAttribArray(v2FAttribute)
        GL.glVertexAttribPointer(v2FAttribute, 1, GL.GL_INT, GL.GL_FALSE, 0, None)


    def defaultFiberValidator(self):
        return np.ones(self.validFiberTexDims[1]*self.validFiberTexDims[0], dtype=np.uint8)


    def configFiberValidator(self):
        maxTexDim = GL.glGetIntegerv(GL.GL_MAX_TEXTURE_SIZE)

        if maxTexDim*maxTexDim < self.curvescount:
            raise ValueError('Fiber data set to big for GPU, need to change rander mode...')
        self.validFiberTexDims = (maxTexDim, int(math.ceil(self.curvescount/maxTexDim)))
        # self. validFiberTexDims = findIntegersMultiplierFor(self.curvescount, maxTexDim)

        self.fiberValidator = self.defaultFiberValidator()

        self.validFiberTexture = GL.glGenTextures(1)

        GL.glActiveTexture(GL.GL_TEXTURE1)

        GL.glBindTexture(GL.GL_TEXTURE_2D, self.validFiberTexture)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_BORDER)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_BORDER)

        bgColor = np.array([1,1,1,1], dtype=np.float32)
        GL.glTexParameterfv(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_BORDER_COLOR, bgColor)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)

        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_R8UI, self.validFiberTexDims[0], self.validFiberTexDims[1], 0, GL.GL_RED_INTEGER, GL.GL_UNSIGNED_BYTE, self.fiberValidator)



    def setAlpha(self, newAlpha):
        self.alpha = newAlpha


    @propagateToChildren
    @drawable
    @config
    def draw(self):

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.colorTableTexture)

        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.validFiberTexture)

        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('colorTable'), 0)
        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('fiberValidator'), 1)

        # Collided fiber
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation("alpha"), 5.0)
        GL.glDrawElements(GL.GL_LINE_STRIP, self.eboSize, GL.GL_UNSIGNED_INT, None)

        # Transparent fiber
        if self.alpha >= 0.01:
            GL.glEnable(GL.GL_BLEND)
            GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
            GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation("alpha"), self.alpha)
            GL.glDrawElements(GL.GL_LINE_STRIP, self.eboSize, GL.GL_UNSIGNED_INT, None)
            GL.glDisable(GL.GL_BLEND)

        self.boundingbox.draw()

    @staticmethod
    def createProgram():
        ''' Anonymous function.
        Default shader program for segmentation.
        It creates the shader programs for this specific class and returns the handler.
        '''

        vertexGLSL = [str(_s_vs)]
        fragmentGLSL = [str(_sfs_vs)]
        geometryGLSL = [str(_s_gs)]
        return [Shader(vertexGLSL, fragmentGLSL, geometryGLSL)]


    def exportFile(self, outfile):
        extension = outfile.split('.')[-1]

        # Bundle file
        if extension == 'bundles':
            self.exportBundleFile(outfile)

        # Trk file
        elif extension == 'trk':
            self.exportTrkFile(outfile)

        # Unsopported file
        else:
            raise TypeError('Unsupported file extension: ', extension)


    # Works only for classes that use fiberValidator array
    def exportBundleFile(self, outfile):
        if self.fiberValidator[:self.curvescount].sum() == 0:
            return

        bundleCount = np.zeros(len(self.bundlesName), dtype=np.int32)

        ROISegmentationExportBundlesdata(
            outfile+'data',
            len(self.bundlesName),
            self.bundlesStart,
            self.fiberSizes,
            bundleCount,
            self.points,
            self.fiberValidator)


        # cfuncs.ROISegmentationExportbundlesdata(
        #       (outfile+'data').encode('utf-8'),
        #       len(self.bundlesName),
        #       self.bundlesStart.ctypes.data,
        #       self.fiberSizes.ctypes.data,
        #       bundleCount.ctypes.data,
        #       self.points.ctypes.data,
        #       self.fiberValidator.ctypes.data)

        ncount = (self.fiberValidator[:self.curvescount] != 0).sum()

        # wrtie minf file
        minf = """attributes = {\n    'binary' : 1,\n    'bundles' : %s,\n    'byte_order' : 'DCBA',\n    'curves_count' : %s,\n    'data_file_name' : '*.bundlesdata',\n    'format' : 'bundles_1.0',\n    'space_dimension' : 3\n  }\n"""

        bundlesstr = []
        offset = 0

        for name, bundleSize in zip([self.bundlesName[i] for i in range(len(self.bundlesName)) if bundleCount[i] != 0], bundleCount[bundleCount != 0]):
            bundlesstr.append(name)
            bundlesstr.append(offset)
            offset += bundleSize

        bundlesstr = str(bundlesstr)
        bundlesstr = bundlesstr[0] + ' ' + bundlesstr[1:-1] + ' ' + bundlesstr[-1]

        with open(outfile, 'w') as file:
            file.write( minf % ( bundlesstr, ncount ) )


    def exportTrkFile(self, outfile):
        ''' This method should create a trk file using the vbo, ebo and the segmentation data (not defined yet).
        '''
        raise TypeError('Exporting trk for this segmentation method is not implemented. Class: ', type(self))



from PyQt5 import QtGui, QtWidgets, uic, QtCore
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

_aipsd_ui = files('phybers.fibervis.ui') / 'Segmentations' / 'InPlaceSegmentationDialog.ui'

class InPlaceSegmentationDialog(QtWidgets.QDialog):
    updateObject = QtCore.pyqtSignal(dict)


    def __init__(self, segmentation, parent):
        super().__init__(parent)
        self.ui = uic.loadUi(str(_aipsd_ui), self)

        self.segmentationRef = segmentation

        self.bundleNames = None
        self.bundleSize = None
        self.states = []
        self.setBundles(self.segmentationRef.bundlesName, self.segmentationRef.bundlesStart)

        self.ui.tableWidget.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.ui.tableWidget.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.ui.tableWidget.setColumnWidth(0, 150)
        self.ui.tableWidget.horizontalHeader().setStretchLastSection(True)

        self.ui.tableWidget.cellClicked.connect(self.toggleCheck)

        self.ui.checkVisualizeAll.clicked.connect(self.visualizeAll)

        self.ui.percentSlider.valueChanged.connect(self.checkRenderState)

        self.ui.filterLineEdit.textChanged.connect(self.filter)

        self.ui.alphaSlider.valueChanged.connect(self.alphaChanged)

        self.ui.renderButton.clicked.connect(self.callUpdate)
        self.ui.exportBundleButton.clicked.connect(self.exportBundle)

        self.show()


    @QtCore.pyqtSlot(int)
    def toggleCheck(self, row):
        tmpItem = self.ui.tableWidget.item(row,0)
        tmpItem.setCheckState(not tmpItem.checkState())


    @QtCore.pyqtSlot()
    def visualizeAll(self):
        self.ui.tableWidget.cellChanged.disconnect(self.toggleStates)

        visualize = self.ui.checkVisualizeAll.isChecked()

        for i in range(self.ui.tableWidget.rowCount()):
            item = self.ui.tableWidget.item(i,0)
            item.setCheckState(visualize)
            self.states[self.bundleNames.index(item.text())] = visualize

        if self.ui.checkRender.isChecked():
            self.updateObject.emit(self.returnDict())

        self.ui.tableWidget.cellChanged.connect(self.toggleStates)


    @QtCore.pyqtSlot()
    def callUpdate(self):
        self.updateObject.emit(self.returnDict())


    @QtCore.pyqtSlot()
    def checkRenderState(self):
        if self.ui.checkRender.isChecked():
            self.updateObject.emit(self.returnDict())


    # TMP!!!!!!
    @QtCore.pyqtSlot(str)
    def filter(self, text):
        self.ui.tableWidget.cellChanged.disconnect(self.toggleStates)

        self.ui.tableWidget.clearContents()
        self.ui.tableWidget.setRowCount(0)

        row = 0
        for i in range(len(self.bundleNames)):
            if self.bundleNames[i].find(text) != -1:
                self.ui.tableWidget.insertRow(row)
                item = QtWidgets.QTableWidgetItem(self.bundleNames[i])
                item.setCheckState(self.states[i])
                self.ui.tableWidget.setItem(row, 0, item)

                self.ui.tableWidget.setItem(row, 1, QtWidgets.QTableWidgetItem(str(self.bundleSize[i])))
                self.ui.tableWidget.item(row,1).setTextAlignment(QtCore.Qt.AlignRight)
                row = row + 1

        checkStateN = 0

        for i in range(self.ui.tableWidget.rowCount()):
            if self.ui.tableWidget.item(i,0).checkState():
                checkStateN += 1

        if checkStateN == self.ui.tableWidget.rowCount():
            self.ui.checkVisualizeAll.setChecked(True)
        elif checkStateN == 0:
            self.ui.checkVisualizeAll.setChecked(False)

        self.ui.tableWidget.cellChanged.connect(self.toggleStates)


    @QtCore.pyqtSlot(int)
    def toggleStates(self, row):
        item = self.ui.tableWidget.item(row,0)
        tmp = self.bundleNames.index(item.text())
        self.states[tmp] = bool(item.checkState())

        tmp = 0

        for i in range(self.ui.tableWidget.rowCount()):
            tmp += int(self.ui.tableWidget.item(i,0).checkState())

        if tmp == self.ui.tableWidget.rowCount():
            self.ui.checkVisualizeAll.setChecked(True)
        else:
            self.ui.checkVisualizeAll.setChecked(False)

        if self.ui.checkRender.isChecked():
            self.updateObject.emit(self.returnDict())


    @QtCore.pyqtSlot()
    def exportBundle(self):

        # Gets direction of bundle files
        outFile, fileType = QtWidgets.QFileDialog.getSaveFileName(self, "Save Bundle file", "", "Bundle (*.bundles)")#;;Trk file (*.trk)")

        options = {     'reference' : self.segmentationRef,
                                'attributes' : {'exportFile' : [outFile]}}

        self.updateObject.emit(options)


    @QtCore.pyqtSlot()
    def alphaChanged(self):
        alpha = self.alphaSlider.value()/(self.alphaSlider.maximum()+1)

        options = {     'reference' : self.segmentationRef,
                                'attributes': {'setAlpha' : [alpha]}}

        self.updateObject.emit(options)


    def setBundles(self, bundlenames, bundlesize):
        self.bundleNames = bundlenames
        self.bundleSize = bundlesize[1:]-bundlesize[:-1]

        # self.ui.tableWidget.cellChanged.disconnect(self.toggleStates)

        self.ui.tableWidget.clearContents()

        for i in range(len(self.bundleNames)):
            self.states.append(True)

            self.ui.tableWidget.insertRow(i)
            item = QtWidgets.QTableWidgetItem(self.bundleNames[i])
            item.setCheckState(True)
            self.ui.tableWidget.setItem(i,0,item)

            self.ui.tableWidget.setItem(i,1,QtWidgets.QTableWidgetItem(str(self.bundleSize[i])))
            self.ui.tableWidget.item(i,1).setTextAlignment(QtCore.Qt.AlignRight)

        self.ui.checkVisualizeAll.setChecked(True)

        self.ui.tableWidget.cellChanged.connect(self.toggleStates)

    def returnDict(self):
        return {'reference' : self.segmentationRef,
                        'attributes' : {'setPercentage' : [self.ui.percentSlider.value()],
                                                        'setStates' : [self.states]},
                        'segmentSubject' : True}


    def __del__(self):
        print('Must clean befor leaving... INPLACESEGMENTATIONDIALOG', type(self))
        segmentationRef = None
        # super.__del__()


    def update(self):
        pass



# shouldn't notify objects by itself... it should be the glcontext
import numpy as np
from .glm import Quaternion, lookAt


class Camera():
    ro_right = np.frombuffer(np.array([1, 0, 0], np.float32), np.float32)
    ro_up = np.frombuffer(np.array([0, 1, 0], np.float32), np.float32)
    ro_front = np.frombuffer(np.array([0, 0, 1], np.float32), np.float32)
    ro_left = np.frombuffer(np.array([-1, 0, 0], np.float32), np.float32)
    ro_down = np.frombuffer(np.array([0, -1, 0], np.float32), np.float32)
    ro_back = np.frombuffer(np.array([0, 0, -1], np.float32), np.float32)

    def __init__(self, radius):
        # reference to objects that needs to be notify
        self.notifyUpdates2Objects = []

        self.radius = radius
        self.theta = 0.0
        self.phi = 0.0
        self.up = self.ro_up
        self.forward = self.ro_back
        self.right = self.ro_right
        self.center = np.array([0, 0, 0], dtype=np.float32)
        self.eye = self.ro_front * radius
        self.p_speed = 0.5 / radius
        self.r_speed = np.pi / 360
        self.z_speed = 0.05
        self.rotation = Quaternion.identity()
        self.view = self.calculateView()

    def calculateView(self):
        rot_phi = Quaternion.fromAngleAxis(self.phi, self.up)
        rot_theta = Quaternion.fromAngleAxis(self.theta, self.right)
        self.rotation = rot_phi * rot_theta
        eye_vector = self.forward * self.radius
        self.eye = self.center - self.rotation.rotateVector(eye_vector)
        if np.round(np.abs(self.theta) - (np.pi / 2), decimals=10) == 0:
            up = rot_phi.rotateVector(self.forward * -np.sign(self.theta))
            self.view = lookAt(self.eye, self.center, up)
        else:
            self.view = lookAt(self.eye, self.center, self.up)
        for object2Notify in self.notifyUpdates2Objects:
            object2Notify.updateCamera(self.eye)
        return self.view

    def orbit(self, dx, dy):
        self.phi = (self.phi - dx * self.r_speed) % (np.pi * 2)
        self.theta = np.clip(self.theta - dy * self.r_speed, -np.pi / 2, np.pi / 2)
        return self.calculateView()

    def spin(self, theta):
        axis = self.center - self.eye
        rotation = Quaternion.fromAngleAxis(theta * self.r_speed, axis)
        self.up = rotation.rotateVector(self.up)
        self.forward = rotation.rotateVector(self.forward)
        self.right = rotation.rotateVector(self.right)
        return self.calculateView()

    def panning(self, dx, dy, dz=0):
        up = self.rotation.rotateVector(self.up)
        right = self.rotation.rotateVector(self.right)
        self.center += (right * -dx + up * dy) * self.p_speed * self.radius
        if dz:
            forward = self.rotation.rotateVector(self.forward)
            self.center += forward * dz * self.p_speed * self.radius
        return self.calculateView()

    def rotate(self, dx, dy):
        m = np.sqrt(dx**2 + dy**2)
        if m > 0:
            axis = [-dy/m, -dx/m, 0]
            theta = np.pi * self.r_speed * m / 180.0
            rotation = Quaternion.fromAngleAxis(theta, axis)
            self.center = self.eye + rotation.rotateVector(self.center - self.eye)
            self.rotation *= rotation
            self.view = self.calculateView()

    def focus(self, coords, radius=None, min_r=None):
        self.center = np.asarray(coords)
        if radius is not None:
            self.radius = radius
        elif min_r is not None:
            self.radius = max(min_r, self.radius)
        return self.calculateView()

    def tiltAxisFromScreen(self):
        return self.rotation.rotateVector(-self.forward)

    def axisAndAngleFromScreen(self, dx, dy):
        up = self.rotation.rotateVector(self.up)
        right = self.rotation.rotateVector(self.right)
        magnitude = np.sqrt(dx**2 + dy**2)
        axis = up * dx / magnitude - right * dy / magnitude
        return axis, magnitude

    def vectorFromScreen(self, dx, dy):
        up = self.rotation.rotateVector(self.up)
        right = self.rotation.rotateVector(self.right)
        translateV = (dx * right + dy * up) * self.p_speed * self.radius
        return translateV

    def zooming(self, zoom):
        self.radius = max(self.radius + self.z_speed * -zoom, 1.0)
        return self.calculateView()

    def changeRotationSpeed(self, speed):
        self.r_speed = speed

    def changeZoomSpeed(self, speed):
        self.z_speed = speed

    def frontView(self):
        self.up = self.ro_up
        self.forward = self.ro_back
        self.right = self.ro_right
        self.phi = np.pi * 0
        self.theta = 0
        return self.calculateView()

    def backView(self):
        self.up = self.ro_up
        self.forward = self.ro_back
        self.right = self.ro_right
        self.phi = np.pi * 1.0
        self.theta = 0
        return self.calculateView()

    def leftView(self):
        self.up = self.ro_up
        self.forward = self.ro_back
        self.right = self.ro_right
        self.phi = np.pi * 1.5
        self.theta = 0
        return self.calculateView()

    def rightView(self):
        self.up = self.ro_up
        self.forward = self.ro_back
        self.right = self.ro_right
        self.phi = np.pi * 0.5
        self.theta = 0
        return self.calculateView()

    def topView(self):
        self.up = self.ro_back
        self.forward = self.ro_down
        self.right = self.ro_right
        self.phi = 0
        self.theta = 0
        return self.calculateView()

    def bottomView(self):
        self.up = self.ro_back
        self.forward = self.ro_up
        self.right = self.ro_left
        self.phi = 0
        self.theta = 0
        return self.calculateView()

    def setObject2Notify(self, object):
        if object not in self.notifyUpdates2Objects:
            self.notifyUpdates2Objects.append(object)

    def removeObject2Notify(self, object):
        if object in self.notifyUpdates2Objects:
            self.notifyUpdates2Objects.remove(object)
